import sys
import time
import os
import json
import stat

from PyQt5 import QtCore, QtGui, uic, QtWidgets
from PyQt5.QtCore import QCoreApplication as QC

from opensnitch.config import Config
from opensnitch.nodes import Nodes
from opensnitch.database import Database
from opensnitch.utils import Message, QuickHelp, Icons, languages
from opensnitch.utils.xdg import Autostart
from opensnitch.utils.themes import Themes
from opensnitch.notifications import DesktopNotifications
from opensnitch.rules import DefaultRulesPath

from opensnitch import auth
import opensnitch.proto as proto
ui_pb2, ui_pb2_grpc = proto.import_()

DIALOG_UI_PATH = "%s/../res/preferences.ui" % os.path.dirname(sys.modules[__name__].__file__)
class PreferencesDialog(QtWidgets.QDialog, uic.loadUiType(DIALOG_UI_PATH)[0]):

    LOG_TAG = "[Preferences] "
    _notification_callback = QtCore.pyqtSignal(str, ui_pb2.NotificationReply)
    saved = QtCore.pyqtSignal()

    TAB_POPUPS = 0
    TAB_UI = 1
    TAB_RULES = 2
    TAB_NODES = 3
    TAB_DB = 4

    NODE_PAGE_GENERAL = 0
    NODE_PAGE_LOGGING = 1
    NODE_PAGE_AUTH = 2

    SUM = 1
    REST = 0

    AUTH_SIMPLE = 0
    AUTH_TLS_SIMPLE = 1
    AUTH_TLS_MUTUAL = 2

    NODE_AUTH = {
        AUTH_SIMPLE: auth.Simple,
        AUTH_TLS_SIMPLE: auth.TLSSimple,
        AUTH_TLS_MUTUAL: auth.TLSMutual
    }
    NODE_AUTH_VERIFY = {
        0: auth.NO_CLIENT_CERT,
        1: auth.REQ_CERT,
        2: auth.REQ_ANY_CERT,
        3: auth.VERIFY_CERT,
        4: auth.REQ_AND_VERIFY_CERT
    }

    def __init__(self, parent=None, appicon=None):
        QtWidgets.QDialog.__init__(self, parent, QtCore.Qt.WindowStaysOnTopHint)

        self._themes = Themes.instance()
        self._saved_theme = ""
        self._restart_msg = QC.translate("preferences", "Restart the GUI in order changes to take effect")
        self._changes_needs_restart = None

        self._cfg = Config.get()
        self._nodes = Nodes.instance()
        self._db = Database.instance()
        self._autostart = Autostart()

        self._notification_callback.connect(self._cb_notification_callback)
        self._notifications_sent = {}
        self._desktop_notifications = DesktopNotifications()

        self.setupUi(self)
        self.setWindowIcon(appicon)

        self.checkDBMaxDays.setEnabled(True)
        self.dbFileButton.setVisible(False)
        self.dbLabel.setVisible(False)
        self.dbType = None

        doubleValidator = QtGui.QDoubleValidator(0, 20, 2, self)
        intValidator = QtGui.QIntValidator(0, 999999, self)
        self.lineUIScreenFactor.setValidator(doubleValidator)
        self.lineNodeMaxEvents.setValidator(intValidator)
        self.lineNodeMaxStats.setValidator(intValidator)
        self.lineNodeFwMonInterval.setValidator(intValidator)

        self.acceptButton.clicked.connect(self._cb_accept_button_clicked)
        self.applyButton.clicked.connect(self._cb_apply_button_clicked)
        self.cancelButton.clicked.connect(self._cb_cancel_button_clicked)
        self.helpButton.clicked.connect(self._cb_help_button_clicked)
        self.popupsCheck.clicked.connect(self._cb_popups_check_toggled)
        self.dbFileButton.clicked.connect(self._cb_file_db_clicked)
        self.cmdTimeoutUp.clicked.connect(lambda: self._cb_cmd_spin_clicked(self.spinUITimeout, self.SUM))
        self.cmdTimeoutDown.clicked.connect(lambda: self._cb_cmd_spin_clicked(self.spinUITimeout, self.REST))
        self.cmdRefreshUIUp.clicked.connect(lambda: self._cb_cmd_spin_clicked(self.spinUIRefresh, self.SUM))
        self.cmdRefreshUIDown.clicked.connect(lambda: self._cb_cmd_spin_clicked(self.spinUIRefresh, self.REST))
        self.cmdUIDensityUp.clicked.connect(lambda: self._cb_cmd_spin_clicked(self.spinUIDensity, self.SUM))
        self.cmdUIDensityDown.clicked.connect(lambda: self._cb_cmd_spin_clicked(self.spinUIDensity, self.REST))
        self.cmdNodeGcUp.clicked.connect(lambda: self._cb_cmd_spin_clicked(self.spinNodeGC, self.SUM))
        self.cmdNodeGcDown.clicked.connect(lambda: self._cb_cmd_spin_clicked(self.spinNodeGC, self.REST))
        self.cmdNodeRulesPath.clicked.connect(self._cb_cmd_node_rulespath_clicked)
        self.cmdDBMaxDaysUp.clicked.connect(lambda: self._cb_cmd_spin_clicked(self.spinDBMaxDays, self.SUM))
        self.cmdDBMaxDaysDown.clicked.connect(lambda: self._cb_cmd_spin_clicked(self.spinDBMaxDays, self.REST))
        self.cmdDBPurgesUp.clicked.connect(lambda: self._cb_cmd_spin_clicked(self.spinDBPurgeInterval, self.SUM))
        self.cmdDBPurgesDown.clicked.connect(lambda: self._cb_cmd_spin_clicked(self.spinDBPurgeInterval, self.REST))
        self.cmdTestNotifs.clicked.connect(self._cb_test_notifs_clicked)
        self.radioSysNotifs.clicked.connect(self._cb_radio_system_notifications)
        self.helpButton.setToolTipDuration(30 * 1000)

        self.comboAuthType.currentIndexChanged.connect(self._cb_combo_auth_type_changed)
        self.comboAuthType.setItemData(PreferencesDialog.AUTH_SIMPLE, auth.Simple)
        self.comboAuthType.setItemData(PreferencesDialog.AUTH_TLS_SIMPLE, auth.TLSSimple)
        self.comboAuthType.setItemData(PreferencesDialog.AUTH_TLS_MUTUAL, auth.TLSMutual)
        self.comboNodeAuthType.setItemData(PreferencesDialog.AUTH_SIMPLE, auth.Simple)
        self.comboNodeAuthType.setItemData(PreferencesDialog.AUTH_TLS_SIMPLE, auth.TLSSimple)
        self.comboNodeAuthType.setItemData(PreferencesDialog.AUTH_TLS_MUTUAL, auth.TLSMutual)
        self.comboNodeAuthVerifyType.setItemData(0, auth.NO_CLIENT_CERT)
        self.comboNodeAuthVerifyType.setItemData(1, auth.REQ_CERT)
        self.comboNodeAuthVerifyType.setItemData(2, auth.REQ_ANY_CERT)
        self.comboNodeAuthVerifyType.setItemData(3, auth.VERIFY_CERT)
        self.comboNodeAuthVerifyType.setItemData(4, auth.REQ_AND_VERIFY_CERT)

        self.comboUIRules.currentIndexChanged.connect(self._cb_combo_uirules_changed)

        # XXX: disable Node duration. It will be removed in the future
        self.comboNodeDuration.setVisible(False)
        self.labelNodeDuration.setVisible(False)

        saveIcon = Icons.new(self, "document-save")
        applyIcon = Icons.new(self, "emblem-default")
        delIcon = Icons.new(self, "edit-delete")
        closeIcon = Icons.new(self, "window-close")
        openIcon = Icons.new(self, "document-open")
        helpIcon = Icons.new(self, "help-browser")
        addIcon = Icons.new(self, "list-add")
        delIcon = Icons.new(self, "list-remove")
        allowIcon = Icons.new(self, "emblem-default")
        denyIcon = Icons.new(self, "emblem-important")
        rejectIcon = Icons.new(self, "window-close")
        self.applyButton.setIcon(applyIcon)
        self.cancelButton.setIcon(closeIcon)
        self.acceptButton.setIcon(saveIcon)
        self.helpButton.setIcon(helpIcon)
        self.dbFileButton.setIcon(openIcon)

        self.cmdTimeoutUp.setIcon(addIcon)
        self.cmdTimeoutDown.setIcon(delIcon)
        self.cmdRefreshUIUp.setIcon(addIcon)
        self.cmdRefreshUIDown.setIcon(delIcon)
        self.cmdUIDensityUp.setIcon(addIcon)
        self.cmdUIDensityDown.setIcon(delIcon)
        self.cmdDBMaxDaysUp.setIcon(addIcon)
        self.cmdDBMaxDaysDown.setIcon(delIcon)
        self.cmdDBPurgesUp.setIcon(addIcon)
        self.cmdDBPurgesDown.setIcon(delIcon)
        self.cmdNodeGcUp.setIcon(addIcon)
        self.cmdNodeGcDown.setIcon(delIcon)

        self.comboUIAction.setItemIcon(Config.ACTION_DENY_IDX, denyIcon)
        self.comboUIAction.setItemIcon(Config.ACTION_ALLOW_IDX, allowIcon)
        self.comboUIAction.setItemIcon(Config.ACTION_REJECT_IDX, rejectIcon)

    def showEvent(self, event):
        super(PreferencesDialog, self).showEvent(event)

        try:
            self._changes_needs_restart = None
            self._settingsSaved = False
            self._reset_status_message()
            self._hide_status_label()
            self.comboNodes.clear()

            self._load_langs()

            self.comboNodeAddress.clear()
            self.comboServerAddr.clear()
            run_path = "/run/user/{0}/opensnitch/".format(os.getuid())
            var_run_path = "/var{0}".format(run_path)
            self.comboNodeAddress.addItem("unix:///tmp/osui.sock")
            self.comboServerAddr.addItem("unix:///tmp/osui.sock")
            if os.path.exists(run_path):
                self.comboNodeAddress.addItem("unix://%s/osui.sock" % run_path)
                self.comboServerAddr.addItem("unix://%s/osui.sock" % run_path)
            if os.path.exists(var_run_path):
                self.comboNodeAddress.addItem("unix://%s/osui.sock" % var_run_path)
                self.comboServerAddr.addItem("unix://%s/osui.sock" % var_run_path)

            self._node_list = self._nodes.get()
            for addr in self._node_list:
                self.comboNodes.addItem(addr)

            if len(self._node_list) == 0:
                self._reset_node_settings()
                self._set_status_message(QC.translate("preferences", "There're no nodes connected"))

            showNodes = len(self._node_list) > 1
            self.comboNodes.setVisible(showNodes)
            self.checkApplyToNodes.setVisible(showNodes)
        except Exception as e:
            print(self.LOG_TAG + "exception loading nodes:", e)

        self._load_settings()

        # connect the signals after loading settings, to avoid firing
        # the signals
        self.comboNodes.currentIndexChanged.connect(self._cb_node_combo_changed)
        self.comboNodeAction.currentIndexChanged.connect(self._cb_node_needs_update)
        #self.comboNodeDuration.currentIndexChanged.connect(self._cb_node_needs_update)
        self.comboNodeMonitorMethod.currentIndexChanged.connect(self._cb_node_needs_update)
        self.comboNodeLogLevel.currentIndexChanged.connect(self._cb_node_needs_update)
        self.comboNodeLogFile.currentIndexChanged.connect(self._cb_node_needs_update)
        self.checkNodeLogUTC.clicked.connect(self._cb_node_needs_update)
        self.checkNodeLogMicro.clicked.connect(self._cb_node_needs_update)
        self.comboNodeAddress.currentTextChanged.connect(self._cb_node_needs_update)
        self.comboServerAddr.currentTextChanged.connect(self._cb_node_needs_update)
        self.checkInterceptUnknown.clicked.connect(self._cb_node_needs_update)
        self.checkApplyToNodes.clicked.connect(self._cb_node_needs_update)
        self.comboNodeAction.currentIndexChanged.connect(self._cb_node_needs_update)
        self.checkNodeAuthSkipVerify.clicked.connect(self._cb_node_needs_update)
        self.comboNodeAuthVerifyType.currentIndexChanged.connect(self._cb_node_needs_update)
        self.enableChecksums.clicked.connect(self._cb_node_needs_update)
        self.checkNodeFlushConns.clicked.connect(self._cb_node_needs_update)
        self.checkNodeBypassQueue.clicked.connect(self._cb_node_needs_update)
        self.spinNodeGC.valueChanged.connect(self._cb_node_needs_update)
        self.lineNodeMaxEvents.textChanged.connect(self._cb_node_needs_update)
        self.lineNodeMaxStats.textChanged.connect(self._cb_node_needs_update)
        self.lineNodeFwMonInterval.textChanged.connect(self._cb_node_needs_update)
        self.lineNodeRulesPath.textChanged.connect(self._cb_node_needs_update)

        self.comboAuthType.currentIndexChanged.connect(self._cb_combo_auth_type_changed)
        self.comboNodeAuthType.currentIndexChanged.connect(self._cb_combo_node_auth_type_changed)

        self.lineCACertFile.textChanged.connect(self._cb_line_certs_changed)
        self.lineCertFile.textChanged.connect(self._cb_line_certs_changed)
        self.lineCertKeyFile.textChanged.connect(self._cb_line_certs_changed)
        self.lineNodeCACertFile.textChanged.connect(self._cb_node_line_certs_changed)
        self.lineNodeCertFile.textChanged.connect(self._cb_node_line_certs_changed)
        self.lineNodeCertKeyFile.textChanged.connect(self._cb_node_line_certs_changed)

        self.lineUIScreenFactor.textChanged.connect(self._cb_ui_screen_factor_changed)
        self.checkUIRules.toggled.connect(self._cb_ui_check_rules_toggled)
        self.checkUIAutoScreen.toggled.connect(self._cb_ui_check_auto_scale_toggled)
        self.comboUITheme.currentIndexChanged.connect(self._cb_combo_themes_changed)
        self.spinUIDensity.valueChanged.connect(self._cb_spin_uidensity_changed)

        self.comboDBType.currentIndexChanged.connect(self._cb_db_type_changed)
        self.checkDBMaxDays.toggled.connect(self._cb_db_max_days_toggled)
        self.checkDBJrnlWal.toggled.connect(self._cb_db_jrnl_wal_toggled)

        # True when any node option changes
        self._node_needs_update = False

    def show_node_prefs(self, addr):
        self.show()
        self.comboNodes.setCurrentText(addr)
        self.tabWidget.setCurrentIndex(self.TAB_NODES)

    def _load_langs(self):
        try:
            self.comboUILang.clear()
            self.comboUILang.blockSignals(True)
            self.comboUILang.addItem(QC.translate("preferences", "System default"), "")
            langs, langNames = languages.get_all()
            for idx, lang in enumerate(langs):
                self.comboUILang.addItem(langNames[idx].capitalize(), langs[idx])
            self.comboUILang.blockSignals(False)
        except Exception as e:
            print(self.LOG_TAG + "exception loading languages:", e)

    def _load_themes(self):
        self.comboUITheme.blockSignals(True)
        theme_idx, self._saved_theme, theme_density = self._themes.get_saved_theme()

        self.labelThemeError.setVisible(False)
        self.labelThemeError.setText("")
        self.comboUITheme.clear()
        self.comboUITheme.addItem(QC.translate("preferences", "System"), "System")
        if self._themes.available():
            themes = self._themes.list_themes()
            for t in themes:
                self.comboUITheme.addItem(os.path.basename(t), t)
        else:
            self._saved_theme = ""
            self.labelThemeError.setStyleSheet('color: red')
            self.labelThemeError.setVisible(True)
            self.labelThemeError.setText(QC.translate("preferences", "Themes not available. Install qt-material: pip3 install qt-material"))

        self.comboUITheme.setCurrentIndex(theme_idx)
        self._show_ui_density_widgets(theme_idx)
        try:
            self.spinUIDensity.setValue(int(theme_density))
        except Exception as e:
            print("load_theme() invalid theme density scale:", theme_density, ":", e)

        self.comboUITheme.blockSignals(False)

    def _load_settings(self):
        self._default_action = self._cfg.getInt(self._cfg.DEFAULT_ACTION_KEY)
        self._default_target = self._cfg.getInt(self._cfg.DEFAULT_TARGET_KEY, 0)
        self._default_timeout = self._cfg.getInt(self._cfg.DEFAULT_TIMEOUT_KEY, Config.DEFAULT_TIMEOUT)
        self._disable_popups = self._cfg.getBool(self._cfg.DEFAULT_DISABLE_POPUPS)

        if self._cfg.hasKey(self._cfg.DEFAULT_DURATION_KEY):
            self._default_duration = self._cfg.getInt(self._cfg.DEFAULT_DURATION_KEY)
        else:
            self._default_duration = self._cfg.DEFAULT_DURATION_IDX

        self.comboUIDuration.setCurrentIndex(self._default_duration)
        self.comboUIDialogPos.setCurrentIndex(self._cfg.getInt(self._cfg.DEFAULT_POPUP_POSITION))
        self.comboUIAction.setCurrentIndex(self._default_action)
        self.comboUITarget.setCurrentIndex(self._default_target)
        self.spinUITimeout.setValue(self._default_timeout)
        self.spinUITimeout.setEnabled(not self._disable_popups)
        self.popupsCheck.setChecked(self._disable_popups)

        self.showAdvancedCheck.setChecked(self._cfg.getBool(self._cfg.DEFAULT_POPUP_ADVANCED))
        self.dstIPCheck.setChecked(self._cfg.getBool(self._cfg.DEFAULT_POPUP_ADVANCED_DSTIP))
        self.dstPortCheck.setChecked(self._cfg.getBool(self._cfg.DEFAULT_POPUP_ADVANCED_DSTPORT))
        self.uidCheck.setChecked(self._cfg.getBool(self._cfg.DEFAULT_POPUP_ADVANCED_UID))
        self.checkSum.setChecked(self._cfg.getBool(self._cfg.DEFAULT_POPUP_ADVANCED_CHECKSUM))

        self.comboUIRules.blockSignals(True)
        self.comboUIRules.setCurrentIndex(self._cfg.getInt(self._cfg.DEFAULT_IGNORE_TEMPORARY_RULES))
        self.checkUIRules.setChecked(self._cfg.getBool(self._cfg.DEFAULT_IGNORE_RULES))
        self.comboUIRules.setEnabled(self._cfg.getBool(self._cfg.DEFAULT_IGNORE_RULES))

        #self._set_rules_duration_filter()

        self._cfg.setRulesDurationFilter(
            self._cfg.getBool(self._cfg.DEFAULT_IGNORE_RULES),
            self._cfg.getInt(self._cfg.DEFAULT_IGNORE_TEMPORARY_RULES)
        )
        self.comboUIRules.blockSignals(False)

         # by default, if no configuration exists, enable notifications.
        self.groupNotifs.setChecked(self._cfg.getBool(Config.NOTIFICATIONS_ENABLED, True))
        self.radioSysNotifs.setChecked(
            True if self._cfg.getInt(Config.NOTIFICATIONS_TYPE) == Config.NOTIFICATION_TYPE_SYSTEM and self._desktop_notifications.is_available() == True else False
        )
        self.radioQtNotifs.setChecked(
            True if self._cfg.getInt(Config.NOTIFICATIONS_TYPE) == Config.NOTIFICATION_TYPE_QT or self._desktop_notifications.is_available() == False else False
        )

        ## db
        self.dbType = self._cfg.getInt(self._cfg.DEFAULT_DB_TYPE_KEY)
        self.comboDBType.setCurrentIndex(self.dbType)
        if self.comboDBType.currentIndex() != Database.DB_TYPE_MEMORY:
            self.dbFileButton.setVisible(True)
            self.dbLabel.setVisible(True)
            self.dbLabel.setText(self._cfg.getSettings(self._cfg.DEFAULT_DB_FILE_KEY))
        dbMaxDays = self._cfg.getInt(self._cfg.DEFAULT_DB_MAX_DAYS, 1)
        dbJrnlWal = self._cfg.getBool(self._cfg.DEFAULT_DB_JRNL_WAL)
        dbPurgeInterval = self._cfg.getInt(self._cfg.DEFAULT_DB_PURGE_INTERVAL, 5)
        self._enable_db_cleaner_options(self._cfg.getBool(Config.DEFAULT_DB_PURGE_OLDEST), dbMaxDays)
        self._enable_db_jrnl_wal(self._cfg.getBool(Config.DEFAULT_DB_PURGE_OLDEST), dbJrnlWal)
        self.spinDBMaxDays.setValue(dbMaxDays)
        self.spinDBPurgeInterval.setValue(dbPurgeInterval)

        self._load_themes()
        self._load_node_settings()
        self._load_ui_settings()

    def _load_ui_settings(self):
        self._ui_refresh_interval = self._cfg.getInt(self._cfg.STATS_REFRESH_INTERVAL, 0)
        self.spinUIRefresh.setValue(self._ui_refresh_interval)

        saved_lang = self._cfg.getSettings(Config.DEFAULT_LANGUAGE)
        if saved_lang:
            saved_langname = self._cfg.getSettings(Config.DEFAULT_LANGNAME)
            self.comboUILang.blockSignals(True)
            self.comboUILang.setCurrentText(saved_langname)
            self.comboUILang.blockSignals(False)

        auto_scale = self._cfg.getBool(Config.QT_AUTO_SCREEN_SCALE_FACTOR, default_value=True)
        screen_factor = self._cfg.getSettings(Config.QT_SCREEN_SCALE_FACTOR)
        if screen_factor is None or screen_factor == "":
            screen_factor = "1"
        self.lineUIScreenFactor.setText(screen_factor)
        self.checkUIAutoScreen.blockSignals(True)
        self.checkUIAutoScreen.setChecked(auto_scale)
        self.checkUIAutoScreen.blockSignals(False)
        self._show_ui_scalefactor_widgets(auto_scale)

        qt_platform = self._cfg.getSettings(Config.QT_PLATFORM_PLUGIN)
        if qt_platform is not None and qt_platform != "":
            self.comboUIQtPlatform.setCurrentText(qt_platform)

        self.checkAutostart.setChecked(self._autostart.isEnabled())

        maxmsgsize = self._cfg.getSettings(Config.DEFAULT_SERVER_MAX_MESSAGE_LENGTH)
        if maxmsgsize:
            self.comboGrpcMsgSize.setCurrentText(maxmsgsize)
        else:
            self.comboGrpcMsgSize.setCurrentIndex(0)

        server_addr = self._cfg.getSettings(Config.DEFAULT_SERVER_ADDR)
        if server_addr == "" or server_addr == None:
            server_addr = self.comboServerAddr.itemText(0)
        self.comboServerAddr.setCurrentText(server_addr)

        self.lineCACertFile.setText(self._cfg.getSettings(Config.AUTH_CA_CERT))
        self.lineCertFile.setText(self._cfg.getSettings(Config.AUTH_CERT))
        self.lineCertKeyFile.setText(self._cfg.getSettings(Config.AUTH_CERTKEY))
        authtype_idx = self.comboAuthType.findData(self._cfg.getSettings(Config.AUTH_TYPE))
        if authtype_idx <= 0:
            authtype_idx = 0
            self.lineCACertFile.setEnabled(False)
            self.lineCertFile.setEnabled(False)
            self.lineCertKeyFile.setEnabled(False)
        self.comboAuthType.setCurrentIndex(authtype_idx)

        self._load_ui_columns_config()

    def _load_node_settings(self):
        addr = self.comboNodes.currentText()
        if addr == "":
            return

        try:
            node_data = self._node_list[addr]['data']
            self.labelNodeVersion.setText(node_data.version)
            self.labelNodeName.setText(node_data.name)
            self.comboNodeLogLevel.setCurrentIndex(node_data.logLevel)

            node_config = json.loads(node_data.config)
            self.comboNodeAction.setCurrentText(node_config['DefaultAction'])
            #self.comboNodeDuration.setCurrentText(node_config['DefaultDuration'])
            self.comboNodeMonitorMethod.setCurrentText(node_config['ProcMonitorMethod'])
            self.checkInterceptUnknown.setChecked(node_config['InterceptUnknown'])
            self.comboNodeLogLevel.setCurrentIndex(int(node_config['LogLevel']))

            if node_config.get('LogUTC') == None:
                node_config['LogUTC'] = False
            self.checkNodeLogUTC.setChecked(node_config['LogUTC'])
            if node_config.get('LogMicro') == None:
                node_config['LogMicro'] = False
            self.checkNodeLogMicro.setChecked(node_config['LogMicro'])

            if node_config.get('Server') != None:
                self.comboNodeAddress.setEnabled(True)
                self.comboNodeLogFile.setEnabled(True)

                self.comboNodeAddress.setCurrentText(node_config['Server']['Address'])
                self.comboNodeLogFile.setCurrentText(node_config['Server']['LogFile'])

                self._load_node_auth_settings(node_config['Server'])
            else:
                self.comboNodeAddress.setEnabled(False)
                self.comboNodeLogFile.setEnabled(False)

            rules = node_config.get('Rules')
            if rules == None:
                rules = {}
            if rules.get('EnableChecksums') == None:
                rules['EnableChecksums'] = False
            if rules.get('Path') == None or rules.get('Path') == "":
                rules['Path'] = DefaultRulesPath
            node_config['Rules'] = rules

            self.enableChecksums.setChecked(rules.get('EnableChecksums'))
            self.lineNodeRulesPath.setText(rules.get('Path'))

            internal = node_config.get('Internal')
            if internal == None:
                internal = {}
            if internal.get('FlushConnsOnStart') == None:
                internal['FlushConnsOnStart'] = False
            if internal.get('GCPercent') == None:
                internal['GCPercent'] = 100
            node_config['Internal'] = internal

            self.checkNodeFlushConns.setChecked(internal.get('FlushConnsOnStart'))
            self.spinNodeGC.setValue(internal.get('GCPercent'))

            fwOptions = node_config.get('FwOptions')
            if fwOptions == None:
                fwOptions = {}
            if fwOptions.get('MonitorInterval') == None or fwOptions.get('MonitorInterval') == "":
                fwOptions['MonitorInterval'] = "15s"
            if fwOptions.get('QueueBypass') == None:
                fwOptions['QueueBypass'] = True
            node_config['FwOptions'] = fwOptions

            monInterval = fwOptions['MonitorInterval'][:-1]
            self.lineNodeFwMonInterval.setText(monInterval)
            self.checkNodeBypassQueue.setChecked(not fwOptions.get('QueueBypass'))

            stats = node_config.get('Stats')
            if stats == None:
                stats = {}
            if stats.get('MaxEvents') == None:
                stats['MaxEvents'] = 250
            if stats.get('MaxStats') == None:
                stats['MaxStats'] = 50
            node_config['Stats'] = stats

            self.lineNodeMaxEvents.setText(str(node_config['Stats']['MaxEvents']))
            self.lineNodeMaxStats.setText(str(node_config['Stats']['MaxStats']))

            self._node_list[addr]['data'].config = json.dumps(node_config, indent="    ")

        except Exception as e:
            print(self.LOG_TAG + "exception loading config: ", e)
            self._set_status_error(QC.translate("preferences", "Error loading config: {0}".format(e)))

    def _load_node_config(self, addr):
        """load the config of a node before sending it back to the node"""
        try:
            if self.comboNodeAddress.currentText() == "":
                return None, QC.translate("preferences", "Server address can not be empty")

            node_action = Config.ACTION_DENY
            if self.comboNodeAction.currentIndex() == Config.ACTION_ALLOW_IDX:
                node_action = Config.ACTION_ALLOW
            elif self.comboNodeAction.currentIndex() == Config.ACTION_REJECT_IDX:
                node_action = Config.ACTION_REJECT

            node_duration = Config.DURATION_ONCE

            node_conf = self._nodes.get_node_config(addr)
            if node_conf == None:
                return None, " "
            node_config = json.loads(node_conf)
            node_config['DefaultAction'] = node_action
            node_config['DefaultDuration'] = node_duration
            node_config['ProcMonitorMethod'] = self.comboNodeMonitorMethod.currentText()
            node_config['LogLevel'] = self.comboNodeLogLevel.currentIndex()
            node_config['LogUTC'] = self.checkNodeLogUTC.isChecked()
            node_config['LogMicro'] = self.checkNodeLogMicro.isChecked()
            node_config['InterceptUnknown'] = self.checkInterceptUnknown.isChecked()

            if node_config.get('Server') != None:
                # skip setting Server Address if we're applying the config to all nodes
                node_config['Server']['Address'] = self.comboNodeAddress.currentText()
                node_config['Server']['LogFile'] = self.comboNodeLogFile.currentText()

                cfg = self._save_node_auth_config(node_config['Server'])
                if cfg != None:
                    node_config['Server'] = cfg
            else:
                print(addr, " doesn't have Server item")

            rules = node_config.get('Rules')
            if rules == None:
                rules = {}
            if rules.get('EnableChecksums') == None:
                rules['EnableChecksums'] = False
                self.enableChecksums.setChecked(False)
            if rules.get('Path') == None or rules.get('Path') == "":
                rules['Path'] = DefaultRulesPath
                self.lineNodeRulesPath.setText(DefaultRulesPath)

            rules['EnableChecksums'] = self.enableChecksums.isChecked()
            rules['Path'] = self.lineNodeRulesPath.text()
            node_config['Rules'] = rules

            internal = node_config.get('Internal')
            if internal == None:
                internal = {}
            if internal.get('FlushConnsOnStart') == None:
                internal['FlushConnsOnStart'] = False
                self.checkNodeFlushConns.setChecked(False)
            if internal.get('GCPercent') == None:
                internal['GCPercent'] = 100
                self.spinNodeGC.setValue(100)

            internal['FlushConnsOnStart'] = self.checkNodeFlushConns.isChecked()
            internal['GCPercent'] = self.spinNodeGC.value()
            node_config['Internal'] = internal

            fwOptions = node_config.get('FwOptions')
            if fwOptions == None:
                fwOptions = {}
            if fwOptions.get('MonitorInterval') == None:
                fwOptions['MonitorInterval'] = "15s"
            if fwOptions.get('QueueBypass') == None:
                fwOptions['QueueBypass'] = True
            node_config['FwOptions'] = fwOptions

            fwOptions['QueueBypass'] = not self.checkNodeBypassQueue.isChecked()
            fwOptions['MonitorInterval'] = self.lineNodeFwMonInterval.text() + "s"

            stats = node_config.get('Stats')
            if stats == None:
                stats = {}
            if stats.get('MaxEvents') == None:
                stats['MaxEvents'] = 250
                self.lineNodeMaxEvents.setText("250")
            if stats.get('MaxStats') == None:
                stats['MaxStats'] = 50
                self.lineNodeMaxStats.setText("50")

            stats['MaxEvents'] = int(self.lineNodeMaxEvents.text())
            stats['MaxStats'] = int(self.lineNodeMaxStats.text())
            node_config['Stats'] = stats

            return json.dumps(node_config, indent="    "), None
        except Exception as e:
            print(self.LOG_TAG + "exception loading node config on %s: " % addr, e)
            self._set_status_error(QC.translate("preferences", "Error loading node config: {0}".format(e)))

        return None, QC.translate("preferences", "Error loading {0} configuration").format(addr)

    def _load_node_auth_settings(self, config):
        try:
            if config == None:
                return

            auth = config.get('Authentication')
            authtype_idx = 0
            if auth != None:
                if auth.get('Type') != None:
                    authtype_idx = self.comboNodeAuthType.findData(auth['Type'])
            else:
                config['Authentication'] = {}
                auth = config.get('Authentication')

            self.lineNodeCACertFile.setEnabled(authtype_idx >= 0)
            self.lineNodeServerCertFile.setEnabled(authtype_idx >= 0)
            self.lineNodeCertFile.setEnabled(authtype_idx >= 0)
            self.lineNodeCertKeyFile.setEnabled(authtype_idx >= 0)

            tls = auth.get('TLSOptions')
            if tls != None and authtype_idx >= 0:
                if tls.get('CACert') != None:
                    self.lineNodeCACertFile.setText(tls['CACert'])
                if tls.get('ServerCert') != None:
                    self.lineNodeServerCertFile.setText(tls['ServerCert'])
                if tls.get('ClientCert') != None:
                    self.lineNodeCertFile.setText(tls['ClientCert'])
                if tls.get('ClientKey') != None:
                    self.lineNodeCertKeyFile.setText(tls['ClientKey'])
                if tls.get('SkipVerify') != None:
                    self.checkNodeAuthSkipVerify.setChecked(tls['SkipVerify'])

                if tls.get('ClientAuthType') != None:
                    clienttype_idx = self.comboNodeAuthVerifyType.findData(tls['ClientAuthType'])
                    if clienttype_idx >= 0:
                        self.comboNodeAuthVerifyType.setCurrentIndex(clienttype_idx)

            self.comboNodeAuthType.setCurrentIndex(authtype_idx)
            # signals are connected after this method is called
            self._cb_combo_node_auth_type_changed(authtype_idx)
        except Exception as e:
            print("[prefs] load node auth options exception:", e)
            self._set_status_error(QC.translate("preferences", "Error loading node auth config: {0}".format(e)))

    def _save_node_auth_config(self, config):
        try:
            auth = config.get('Authentication')
            if auth == None:
                auth = {}

            auth['Type'] = self.NODE_AUTH[self.comboNodeAuthType.currentIndex()]
            tls = auth.get('TLSOptions')
            if tls == None:
                tls = {}

            tls['CACert'] = self.lineNodeCACertFile.text()
            tls['ServerCert'] = self.lineNodeServerCertFile.text()
            tls['ClientCert'] = self.lineNodeCertFile.text()
            tls['ClientKey'] = self.lineNodeCertKeyFile.text()
            tls['SkipVerify'] = self.checkNodeAuthSkipVerify.isChecked()
            tls['ClientAuthType'] = self.NODE_AUTH_VERIFY[self.comboNodeAuthVerifyType.currentIndex()]
            auth['TLSOptions'] = tls
            config['Authentication'] = auth

            return config
        except Exception as e:
            print("[prefs] node auth options exception:", e)
            self._set_status_error(str(e))
            return None

    def _load_ui_columns_config(self):
        cols = self._cfg.getSettings(Config.STATS_SHOW_COLUMNS)
        if cols == None:
            return

        for c in range(13):
            checked = str(c) in cols

            if c == 0:
                self.checkHideTime.setChecked(checked)
            elif c == 1:
                self.checkHideNode.setChecked(checked)
            elif c == 2:
                self.checkHideAction.setChecked(checked)
            elif c == 3:
                self.checkHideSrcPort.setChecked(checked)
            elif c == 4:
                self.checkHideSrcIP.setChecked(checked)
            elif c == 5:
                self.checkHideDstIP.setChecked(checked)
            elif c == 6:
                self.checkHideDstHost.setChecked(checked)
            elif c == 7:
                self.checkHideDstPort.setChecked(checked)
            elif c == 8:
                self.checkHideProto.setChecked(checked)
            elif c == 9:
                self.checkHideUID.setChecked(checked)
            elif c == 10:
                self.checkHidePID.setChecked(checked)
            elif c == 11:
                self.checkHideProc.setChecked(checked)
            elif c == 12:
                self.checkHideCmdline.setChecked(checked)
            elif c == 13:
                self.checkHideRule.setChecked(checked)

    def _reset_node_settings(self):
        self.comboNodeAction.setCurrentIndex(0)
        #self.comboNodeDuration.setCurrentIndex(0)
        self.comboNodeMonitorMethod.setCurrentIndex(0)
        self.checkInterceptUnknown.setChecked(False)
        self.comboNodeLogLevel.setCurrentIndex(0)
        self.checkNodeLogUTC.setChecked(True)
        self.checkNodeLogMicro.setChecked(False)
        self.labelNodeName.setText("")
        self.labelNodeVersion.setText("")
        self.comboNodeAuthType.setCurrentIndex(self.AUTH_SIMPLE)
        self.lineNodeCACertFile.setText("")
        self.lineNodeServerCertFile.setText("")
        self.lineNodeCertFile.setText("")
        self.lineNodeCertKeyFile.setText("")
        self.checkNodeAuthSkipVerify.setChecked(False)
        self.comboNodeAuthVerifyType.setCurrentIndex(0)
        self._cb_combo_node_auth_type_changed(0)

    def _save_settings(self):
        self._reset_status_message()
        ui_saved = self._save_ui_config()
        db_saved = self._save_db_config()
        nodes_saved = self._save_nodes_config()

        if ui_saved and db_saved and nodes_saved:
            self._set_status_successful(QC.translate("preferences", "Configuration applied."))
        self.saved.emit()
        self._settingsSaved = True
        self._needs_restart()

    def _save_db_config(self):
        dbtype = self.comboDBType.currentIndex()
        db_name = self._cfg.getSettings(self._cfg.DEFAULT_DB_FILE_KEY)

        if self.dbLabel.text() != "" and \
                (self.comboDBType.currentIndex() != self.dbType or db_name != self.dbLabel.text()):
            self._changes_needs_restart = QC.translate("preferences", "DB type changed")

        if self.comboDBType.currentIndex() != Database.DB_TYPE_MEMORY:
            if self.dbLabel.text() != "":
                db_name = self.dbLabel.text()
            else:
                Message.ok(
                    QC.translate("preferences", "Warning"),
                    QC.translate("preferences", "You must select a file for the database<br>or choose \"In memory\" type."),
                    QtWidgets.QMessageBox.Warning)
                self.dbLabel.setText("")
                return False
        else:
            db_name = Database.DB_IN_MEMORY

        self._cfg.setSettings(Config.DEFAULT_DB_FILE_KEY, db_name)
        self._cfg.setSettings(Config.DEFAULT_DB_TYPE_KEY, dbtype)
        self._cfg.setSettings(Config.DEFAULT_DB_PURGE_OLDEST, bool(self.checkDBMaxDays.isChecked()))
        self._cfg.setSettings(Config.DEFAULT_DB_MAX_DAYS, int(self.spinDBMaxDays.value()))
        self._cfg.setSettings(Config.DEFAULT_DB_PURGE_INTERVAL, int(self.spinDBPurgeInterval.value()))
        self._cfg.setSettings(Config.DEFAULT_DB_JRNL_WAL, bool(self.checkDBJrnlWal.isChecked()))
        self.dbType = self.comboDBType.currentIndex()

        return True

    def _save_ui_config(self):
        try:
            self._save_ui_columns_config()

            maxmsgsize = self.comboGrpcMsgSize.currentText()
            mmsize_saved = self._cfg.getSettings(Config.DEFAULT_SERVER_MAX_MESSAGE_LENGTH)
            if maxmsgsize != "" and mmsize_saved != maxmsgsize:
                self._cfg.setSettings(Config.DEFAULT_SERVER_MAX_MESSAGE_LENGTH, maxmsgsize.replace(" ", ""))
                self._changes_needs_restart = QC.translate("preferences", "Server options changed")

            savedauthtype = self._cfg.getSettings(Config.AUTH_TYPE)
            authtype = self.comboAuthType.itemData(self.comboAuthType.currentIndex())
            cacert = self._cfg.getSettings(Config.AUTH_CA_CERT)
            cert = self._cfg.getSettings(Config.AUTH_CERT)
            certkey = self._cfg.getSettings(Config.AUTH_CERTKEY)
            if not self._validate_certs():
                return False

            server_addr = self._cfg.getSettings(Config.DEFAULT_SERVER_ADDR)
            if self.comboServerAddr.currentText() != server_addr:
                self._cfg.setSettings(Config.DEFAULT_SERVER_ADDR, self.comboServerAddr.currentText())
                self._changes_needs_restart = QC.translate("preferences", "Server address changed")

            if savedauthtype != authtype or self.lineCertFile.text() != cert or \
                    self.lineCertKeyFile.text() != certkey or self.lineCACertFile.text() != cacert:
                self._changes_needs_restart = QC.translate("preferences", "Certificates changed")
            self._cfg.setSettings(Config.AUTH_TYPE, authtype)
            self._cfg.setSettings(Config.AUTH_CA_CERT, self.lineCACertFile.text())
            self._cfg.setSettings(Config.AUTH_CERT, self.lineCertFile.text())
            self._cfg.setSettings(Config.AUTH_CERTKEY, self.lineCertKeyFile.text())

            selected_lang = self.comboUILang.itemData(self.comboUILang.currentIndex())
            saved_lang = self._cfg.getSettings(Config.DEFAULT_LANGUAGE)
            saved_lang = "" if saved_lang is None else saved_lang
            if saved_lang != selected_lang:
                languages.save(self._cfg, selected_lang)
                self._changes_needs_restart = QC.translate("preferences", "Language changed")

            self._cfg.setSettings(self._cfg.DEFAULT_IGNORE_TEMPORARY_RULES, int(self.comboUIRules.currentIndex()))
            self._cfg.setSettings(self._cfg.DEFAULT_IGNORE_RULES, bool(self.checkUIRules.isChecked()))
            #self._set_rules_duration_filter()
            self._cfg.setRulesDurationFilter(
                bool(self.checkUIRules.isChecked()),
                int(self.comboUIRules.currentIndex())
            )
            if self.checkUIRules.isChecked():
                self._nodes.delete_rule_by_field(Config.DURATION_FIELD, Config.RULES_DURATION_FILTER)

            self._cfg.setSettings(self._cfg.STATS_REFRESH_INTERVAL, int(self.spinUIRefresh.value()))
            self._cfg.setSettings(self._cfg.DEFAULT_ACTION_KEY, self.comboUIAction.currentIndex())
            self._cfg.setSettings(self._cfg.DEFAULT_DURATION_KEY, int(self.comboUIDuration.currentIndex()))
            self._cfg.setSettings(self._cfg.DEFAULT_TARGET_KEY, self.comboUITarget.currentIndex())
            self._cfg.setSettings(self._cfg.DEFAULT_TIMEOUT_KEY, self.spinUITimeout.value())
            self._cfg.setSettings(self._cfg.DEFAULT_DISABLE_POPUPS, bool(self.popupsCheck.isChecked()))
            self._cfg.setSettings(self._cfg.DEFAULT_POPUP_POSITION, int(self.comboUIDialogPos.currentIndex()))

            self._cfg.setSettings(self._cfg.DEFAULT_POPUP_ADVANCED, bool(self.showAdvancedCheck.isChecked()))
            self._cfg.setSettings(self._cfg.DEFAULT_POPUP_ADVANCED_DSTIP, bool(self.dstIPCheck.isChecked()))
            self._cfg.setSettings(self._cfg.DEFAULT_POPUP_ADVANCED_DSTPORT, bool(self.dstPortCheck.isChecked()))
            self._cfg.setSettings(self._cfg.DEFAULT_POPUP_ADVANCED_UID, bool(self.uidCheck.isChecked()))
            self._cfg.setSettings(self._cfg.DEFAULT_POPUP_ADVANCED_CHECKSUM, bool(self.checkSum.isChecked()))

            self._cfg.setSettings(self._cfg.NOTIFICATIONS_ENABLED, bool(self.groupNotifs.isChecked()))
            self._cfg.setSettings(self._cfg.NOTIFICATIONS_TYPE,
                                int(Config.NOTIFICATION_TYPE_SYSTEM if self.radioSysNotifs.isChecked() else Config.NOTIFICATION_TYPE_QT))

            thm_name = self.get_theme_name()
            self._themes.save_theme(self.comboUITheme.currentIndex(), thm_name, str(self.spinUIDensity.value()))

            qt_platform = self._cfg.getSettings(Config.QT_PLATFORM_PLUGIN)
            if qt_platform != self.comboUIQtPlatform.currentText():
                self._changes_needs_restart = QC.translate("preferences", "Qt platform plugin changed")
            self._cfg.setSettings(Config.QT_PLATFORM_PLUGIN, self.comboUIQtPlatform.currentText())
            self._cfg.setSettings(Config.QT_AUTO_SCREEN_SCALE_FACTOR, bool(self.checkUIAutoScreen.isChecked()))
            self._cfg.setSettings(Config.QT_SCREEN_SCALE_FACTOR, self.lineUIScreenFactor.text())

            if self._themes.available() and \
                    self._saved_theme != "" and \
                    (self.comboUITheme.currentText() == QC.translate("preferences", "System") or \
                     "dark" in self.comboUITheme.currentText()):
                self._changes_needs_restart = QC.translate("preferences", "UI theme changed")

            # this is a workaround for not display pop-ups.
            # see #79 for more information.
            if self.popupsCheck.isChecked():
                self._cfg.setSettings(self._cfg.DEFAULT_TIMEOUT_KEY, 0)

            self._autostart.enable(self.checkAutostart.isChecked())

            return True

        except Exception as e:
            self._set_status_error(str(e))
            return False

    def _save_ui_columns_config(self):
        cols=list()
        if self.checkHideTime.isChecked():
            cols.append("0")
        if self.checkHideNode.isChecked():
            cols.append("1")
        if self.checkHideAction.isChecked():
            cols.append("2")
        if self.checkHideSrcPort.isChecked():
            cols.append("3")
        if self.checkHideSrcIP.isChecked():
            cols.append("4")
        if self.checkHideDstIP.isChecked():
            cols.append("5")
        if self.checkHideDstHost.isChecked():
            cols.append("6")
        if self.checkHideDstPort.isChecked():
            cols.append("7")
        if self.checkHideProto.isChecked():
            cols.append("8")
        if self.checkHideUID.isChecked():
            cols.append("9")
        if self.checkHidePID.isChecked():
            cols.append("10")
        if self.checkHideProc.isChecked():
            cols.append("11")
        if self.checkHideCmdline.isChecked():
            cols.append("12")
        if self.checkHideRule.isChecked():
            cols.append("13")

        self._cfg.setSettings(Config.STATS_SHOW_COLUMNS, cols)

    def _save_nodes_config(self):
        addr = self.comboNodes.currentText()
        if (self._node_needs_update or self.checkApplyToNodes.isChecked()) and addr != "":
            self._set_status_message(QC.translate("preferences", "Saving configuration..."))
            try:
                notif = ui_pb2.Notification(
                        id=int(str(time.time()).replace(".", "")),
                        type=ui_pb2.CHANGE_CONFIG,
                        data="",
                        rules=[])
                if self.checkApplyToNodes.isChecked():
                    for addr in self._nodes.get_nodes():
                        error = self._save_node_config(notif, addr)
                        if error != None:
                            self._set_status_error(error)
                            return False
                else:
                    error = self._save_node_config(notif, addr)
                    if error != None:
                        self._set_status_error(error)
                        return False
            except Exception as e:
                print(self.LOG_TAG + "exception saving config: ", e)
                self._set_status_error(QC.translate("preferences", "Exception saving config: {0}").format(str(e)))
                return False

        elif addr == "":
            self._set_status_message(QC.translate("preferences", "There're no nodes connected"))
            return False

        self._node_needs_update = False

        return True

    def _save_node_config(self, notifObject, addr):
        try:
            if self._nodes.count() == 0:
                return
            self._set_status_message(QC.translate("preferences", "Applying configuration on {0} ...").format(addr))
            notifObject.data, error = self._load_node_config(addr)
            if error != None:
                return error

            # exclude this message if there're more than one node connected
            # XXX: unix:/local is a special name for the node, when the gRPC
            # does not return the correct address of the node.
            if (self.comboNodes.currentText() != "unix:/local" and self.comboNodes.currentText() != self.comboNodeAddress.currentText()) or \
                    self.comboServerAddr.currentText() != self.comboNodeAddress.currentText():
                self._changes_needs_restart = QC.translate("preferences", "Node address changed (update GUI address if needed)")

            self._nodes.save_node_config(addr, notifObject.data)
            nid = self._nodes.send_notification(addr, notifObject, self._notification_callback)
            self._notifications_sent[nid] = notifObject

        except Exception as e:
            print(self.LOG_TAG + "exception saving node config on %s: " % addr, e)
            self._set_status_error(QC.translate("preferences", "Exception saving node config {0}: {1}").format((addr, str(e))))
            return addr + ": " + str(e)

        return None

    def _validate_certs(self):
        try:
            if self.comboAuthType.currentIndex() == PreferencesDialog.AUTH_SIMPLE:
                return True

            if self.comboAuthType.currentIndex() > 0 and (self.lineCertFile.text() == "" or self.lineCertKeyFile.text() == ""):
                raise ValueError(QC.translate("preferences", "Certs fields cannot be empty."))

            if oct(stat.S_IMODE(os.lstat(self.lineCertFile.text()).st_mode)) != "0o600":
                self._set_status_message(
                    QC.translate("preferences", "cert file has excessive permissions, it should have 0600")
                )
            if oct(stat.S_IMODE(os.lstat(self.lineCertFile.text()).st_mode)) != "0o600":
                self._set_status_message(
                    QC.translate("preferences", "cert key file has excessive permissions, it should have 0600")
                )

            if self.comboAuthType.currentIndex() == PreferencesDialog.AUTH_TLS_MUTUAL:
                if oct(stat.S_IMODE(os.lstat(self.lineCACertFile.text()).st_mode)) != "0o600":
                    self._set_status_message(
                        QC.translate("preferences", "CA cert file has excessive permissions, it should have 0600")
                    )

            return True
        except Exception as e:
            self._changes_needs_restart = None
            self._set_status_error("certs error: {0}".format(e))
            return False

    def _needs_restart(self):
        if self._changes_needs_restart:
            Message.ok(self._changes_needs_restart,
                self._restart_msg,
                QtWidgets.QMessageBox.Warning)
            self._changes_needs_restart = None


    def _show_ui_density_widgets(self, idx):
        """show ui density widget only for qt-material themes:
            https://github.com/UN-GCPDS/qt-material?tab=readme-ov-file#density-scale
        """
        hidden = idx == 0
        self.labelUIDensity.setHidden(hidden)
        self.spinUIDensity.setHidden(hidden)
        self.cmdUIDensityUp.setHidden(hidden)
        self.cmdUIDensityDown.setHidden(hidden)

    def _show_ui_scalefactor_widgets(self, show=False):
        self.labelUIScreenFactor.setHidden(show)
        self.lineUIScreenFactor.setHidden(show)

    def _hide_status_label(self):
        self.statusLabel.hide()

    def _show_status_label(self):
        self.statusLabel.show()

    def _set_status_error(self, msg):
        self._show_status_label()
        self.statusLabel.setStyleSheet('color: red')
        self.statusLabel.setText(msg)
        QtWidgets.QApplication.processEvents()

    def _set_status_successful(self, msg):
        self._show_status_label()
        self.statusLabel.setStyleSheet('color: green')
        self.statusLabel.setText(msg)
        QtWidgets.QApplication.processEvents()

    def _set_status_message(self, msg):
        self._show_status_label()
        self.statusLabel.setStyleSheet('color: darkorange')
        self.statusLabel.setText(msg)
        QtWidgets.QApplication.processEvents()

    def _reset_status_message(self):
        self.statusLabel.setText("")
        self._hide_status_label()
        # force widgets repainting
        QtWidgets.QApplication.processEvents()

    def _enable_db_cleaner_options(self, enable, db_max_days):
        self.checkDBMaxDays.setChecked(enable)
        self.spinDBMaxDays.setEnabled(enable)
        self.spinDBPurgeInterval.setEnabled(enable)
        self.labelDBPurgeInterval.setEnabled(enable)
        self.labelDBPurgeDays.setEnabled(enable)
        self.labelDBPurgeMinutes.setEnabled(enable)
        self.cmdDBMaxDaysUp.setEnabled(enable)
        self.cmdDBMaxDaysDown.setEnabled(enable)
        self.cmdDBPurgesUp.setEnabled(enable)
        self.cmdDBPurgesDown.setEnabled(enable)

    def _enable_db_jrnl_wal(self, enable, db_jrnl_wal):
        self.checkDBJrnlWal.setChecked(db_jrnl_wal)
        self.checkDBJrnlWal.setEnabled(enable)

    def get_theme_name(self):
        thm_idx = self.comboUITheme.currentIndex()
        return self.comboUITheme.itemData(thm_idx)

    def _change_theme(self):
        extra_opts = {
            'density_scale': str(self.spinUIDensity.value())
        }
        thm_name = self.get_theme_name()
        self._themes.change_theme(self, thm_name, extra_opts)

    @QtCore.pyqtSlot(str, ui_pb2.NotificationReply)
    def _cb_notification_callback(self, addr, reply):
        #print(self.LOG_TAG, "Config notification received: ", reply.id, reply.code)
        if reply.id in self._notifications_sent:
            if reply.code == ui_pb2.OK:
                self._set_status_successful(QC.translate("preferences", "Configuration applied."))
            else:
                self._set_status_error(QC.translate("preferences", "Error applying configuration: {0}").format(reply.data))

            del self._notifications_sent[reply.id]

    def _cb_line_certs_changed(self, text):
        self._changes_needs_restart = QC.translate("preferences", "Certs changed")

    def _cb_node_line_certs_changed(self, text):
        self._changes_needs_restart = QC.translate("preferences", "Node certs changed")
        self._node_needs_update = True

    def _cb_cmd_node_rulespath_clicked(self):
        rulesdir = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            os.path.expanduser("~"),
            QC.translate("preferences", 'Select a directory containing rules'),
            QtWidgets.QFileDialog.ShowDirsOnly | QtWidgets.QFileDialog.DontResolveSymlinks
        )
        if rulesdir == "":
            return

        self._node_needs_update = True
        self.lineNodeRulesPath.setText(rulesdir)

    def _cb_file_db_clicked(self):
        options = QtWidgets.QFileDialog.Options()
        fileName, _ = QtWidgets.QFileDialog.getSaveFileName(self, "", "","All Files (*)", options=options)
        if fileName:
            self.dbLabel.setText(fileName)

    def _cb_combo_uirules_changed(self, idx):
        self._cfg.setRulesDurationFilter(
            self._cfg.getBool(self._cfg.DEFAULT_IGNORE_RULES),
            idx
            #self._cfg.getInt(self._cfg.DEFAULT_IGNORE_TEMPORARY_RULES)
        )

    def _cb_db_type_changed(self):
        isDBMem = self.comboDBType.currentIndex() == Database.DB_TYPE_MEMORY
        self.dbFileButton.setVisible(not isDBMem)
        self.dbLabel.setVisible(not isDBMem)
        self.checkDBMaxDays.setChecked(self._cfg.getBool(Config.DEFAULT_DB_PURGE_OLDEST))
        self.checkDBJrnlWal.setEnabled(not isDBMem)
        self.checkDBJrnlWal.setChecked(False)

    def _cb_accept_button_clicked(self):
        self.accept()
        if not self._settingsSaved:
            self._save_settings()

    def _cb_apply_button_clicked(self):
        self._save_settings()

    def _cb_cancel_button_clicked(self):
        self.reject()

    def _cb_help_button_clicked(self):
        QuickHelp.show(
            QC.translate("preferences",
                         "Hover the mouse over the texts to display the help<br><br>Don't forget to visit the wiki: <a href=\"{0}\">{0}</a>"
                         ).format(Config.HELP_URL)
        )

    def _cb_popups_check_toggled(self, checked):
        self.spinUITimeout.setEnabled(not checked)
        if not checked:
            self.spinUITimeout.setValue(20)

    def _cb_node_combo_changed(self, index):
        self._load_node_settings()

    def _cb_node_needs_update(self):
        self._node_needs_update = True

    def _cb_ui_check_rules_toggled(self, state):
        self.comboUIRules.setEnabled(state)

    def _cb_combo_themes_changed(self, index):
        self._change_theme()
        self._show_ui_density_widgets(index)

    def _cb_spin_uidensity_changed(self, value):
        self._change_theme()

    def _cb_ui_check_auto_scale_toggled(self, checked):
        self._changes_needs_restart = QC.translate("preferences", "Auto scale option changed")
        self._show_ui_scalefactor_widgets(checked)

    def _cb_ui_screen_factor_changed(self, text):
        self._changes_needs_restart = QC.translate("preferences", "Screen factor option changed")

    def _cb_combo_auth_type_changed(self, index):
        curtype = self.comboAuthType.itemData(self.comboAuthType.currentIndex())
        savedtype = self._cfg.getSettings(Config.AUTH_TYPE)
        if curtype != savedtype:
            self._changes_needs_restart = QC.translate("preferences", "Auth type changed")

        self.lineCACertFile.setEnabled(index == PreferencesDialog.AUTH_TLS_MUTUAL)
        self.lineCertFile.setEnabled(index >= PreferencesDialog.AUTH_TLS_SIMPLE)
        self.lineCertKeyFile.setEnabled(index >= PreferencesDialog.AUTH_TLS_SIMPLE)

    def _cb_combo_node_auth_type_changed(self, index):
        curtype = self.comboNodeAuthType.itemData(self.comboNodeAuthType.currentIndex())
        #savedtype = self._cfg.getSettings(Config.AUTH_TYPE)
        #if curtype != savedtype:
        #    self._changes_needs_restart = QC.translate("preferences", "Auth type changed")

        self.lineNodeCACertFile.setEnabled(index == PreferencesDialog.AUTH_TLS_MUTUAL)
        self.lineNodeServerCertFile.setEnabled(index >= PreferencesDialog.AUTH_TLS_SIMPLE)
        self.lineNodeCertFile.setEnabled(index >= PreferencesDialog.AUTH_TLS_SIMPLE)
        self.lineNodeCertKeyFile.setEnabled(index >= PreferencesDialog.AUTH_TLS_SIMPLE)
        self.checkNodeAuthSkipVerify.setEnabled(index >= PreferencesDialog.AUTH_TLS_SIMPLE)
        self.comboNodeAuthVerifyType.setEnabled(index >= PreferencesDialog.AUTH_TLS_SIMPLE)

        self._node_needs_update = True

    def _cb_db_max_days_toggled(self, state):
        self._enable_db_cleaner_options(state, 1)

    def _cb_db_jrnl_wal_toggled(self, state):
        self._changes_needs_restart = QC.translate("preferences", "DB journal_mode changed")

    def _cb_cmd_spin_clicked(self, spinWidget, operation):
        if operation == self.SUM:
            spinWidget.setValue(spinWidget.value() + 1)
        else:
            spinWidget.setValue(spinWidget.value() - 1)

        if spinWidget == self.popupsCheck:
            enablePopups = spinWidget.value() > 0
            self.popupsCheck.setChecked(not enablePopups)
            self.spinUITimeout.setEnabled(enablePopups)
            self._node_needs_update = True

    def _cb_radio_system_notifications(self):
        if self._desktop_notifications.is_available() == False:
            self.radioSysNotifs.setChecked(False)
            self.radioQtNotifs.setChecked(True)
            self._set_status_error(QC.translate("notifications", "System notifications are not available, you need to install python3-notify2."))
            return

    def _cb_test_notifs_clicked(self):
        try:
            self.cmdTestNotifs.setEnabled(False)
            if self._desktop_notifications.is_available() == False:
                self._set_status_error(QC.translate("notifications", "System notifications are not available, you need to install python3-notify2."))
                return

            if self.radioSysNotifs.isChecked():
                self._desktop_notifications.show("title", "body")
            else:
                pass
        except Exception as e:
            print(self.LOG_TAG + "exception testing notifications:", e)
        finally:
            self.cmdTestNotifs.setEnabled(True)




from google.protobuf import __version__ as protobuf_version
from .enums import *

class Utils():

    @staticmethod
    def isExprPort(value):
        """Return true if the value is valid for a port based rule:
            nft add rule ... tcp dport 22 accept
        """
        return value == Statements.TCP.value or \
                value == Statements.UDP.value or \
                value == Statements.UDPLITE.value or \
                value == Statements.SCTP.value or \
                value == Statements.DCCP.value

    @staticmethod
    def isProtobufSupported():
        """
        The protobuf operations append() and insert() were introduced on 3.8.0 version.
        """
        vparts = protobuf_version.split(".")
        return int(vparts[0]) >= 3 and int(vparts[1]) >= 8



from PyQt5.QtCore import QCoreApplication as QC
from opensnitch.config import Config

class DesktopNotifications():
    """DesktopNotifications display informative pop-ups using the system D-Bus.
    The notifications are handled and configured by the system.

    The notification daemon also decides where to show the notifications, as well
    as how to group them.

    The body of a notification supports markup (if the implementation supports it):
        https://people.gnome.org/~mccann/docs/notification-spec/notification-spec-latest.html#markup
    Basically: <a>, <u>, <b>, <i> and <img>. New lines can be added with the regular \n.

    It also support actions (buttons).

    https://notify2.readthedocs.io/en/latest/
    """

    _cfg = Config.init()

    # list of hints:
    # https://people.gnome.org/~mccann/docs/notification-spec/notification-spec-latest.html#hints
    HINT_DESKTOP_ENTRY = "desktop-entry"
    CATEGORY_NETWORK = "network"

    EXPIRES_DEFAULT = 0
    NEVER_EXPIRES = -1

    URGENCY_LOW = 0
    URGENCY_NORMAL = 1
    URGENCY_CRITICAL = 2

    # must be a string
    ACTION_ID_OPEN = "action-open"
    ACTION_ID_ALLOW = "action-allow"
    ACTION_ID_DENY = "action-deny"

    def __init__(self):
        self.ACTION_OPEN = QC.translate("popups", "Open")
        self.ACTION_ALLOW = QC.translate("popups", "Allow")
        self.ACTION_DENY = QC.translate("popups", "Deny")
        self.IS_LIBNOTIFY_AVAILABLE = True
        self.DOES_SUPPORT_ACTIONS = True
        self.ntf2 = None

        try:
            import notify2
            self.ntf2 = notify2
            mloop = 'glib'

            # First try to initialise the D-Bus connection with the given
            # mainloop.
            # If it fails, we'll try to initialise it without it.
            try:
                self.ntf2.init("opensnitch", mainloop=mloop)
            except Exception:
                self.DOES_SUPPORT_ACTIONS = False
                if self.ntf2 == None:
                    print("DesktopNotifications() notify2 package not available. Try installing python3-notify2, or 'notify2' as your regular user (pip3 install notify2)")
                    return

                self.ntf2.init("opensnitch")

                # usually because dbus mainloop is not initiated, specially
                # with 'qt'
                # FIXME: figure out how to init it, or how to connect to an
                # existing session.
                print("DesktopNotifications(): system doesn't support actions. Available capabilities:")
                print(self.ntf2.get_server_caps())


            # Example: ['actions', 'action-icons', 'body', 'body-markup', 'icon-static', 'persistence', 'sound']
            if ('actions' not in self.ntf2.get_server_caps()):
                self.DOES_SUPPORT_ACTIONS = False

        except Exception as e:
            print("DesktopNotifications not available (install python3-notify2):", e)
            self.IS_LIBNOTIFY_AVAILABLE = False

    def is_available(self):
        return self.IS_LIBNOTIFY_AVAILABLE

    def are_enabled(self):
        return self._cfg.getBool(Config.NOTIFICATIONS_ENABLED, True)

    def support_actions(self):
        """Returns true if the notifications daemon support actions(buttons).
        This depends on 2 factors:
            - If the notification server actually supports it (get_server_caps()).
            - If there's a dbus instance running.
        """
        return self.DOES_SUPPORT_ACTIONS

    def show(self, title, body, icon="dialog-information", urgency=URGENCY_NORMAL, callback=None):
        try:
            ntf = self.ntf2.Notification(title, body, icon)

            ntf.set_urgency(urgency)
            ntf.set_category(self.CATEGORY_NETWORK)
            # used to display our app icon and name.
            # Note: setting this Hint causes some DEs to call opensnitch_ui.desktop file,
            # that as of today, kills and relaunches the current opensnitch-ui process.
            #ntf.set_hint(self.HINT_DESKTOP_ENTRY, "opensnitch_ui")
            if self.DOES_SUPPORT_ACTIONS and callback != None:
                ntf.add_action(self.ACTION_ID_OPEN, self.ACTION_OPEN, callback)
            ntf.show()
        except Exception as e:
            print("[notifications] show() exception:", e)
            raise Exception("[notifications] show() exception:", e)

    # TODO:
    #  - construct a rule with the default configured parameters.
    #  - create a common dialogs/prompt.py:_send_rule(), maybe in utils.py
    def ask(self, connection, timeout, callback):
        c = connection
        title = QC.translate("popups", "New outgoing connection")
        body = c.process_path + "\n"
        body = body + QC.translate("popups", "is connecting to <b>%s</b> on %s port %d") % ( \
            c.dst_host or c.dst_ip,
            c.protocol.upper(),
            c.dst_port )

        ntf = self.ntf2.Notification(title, body, "dialog-warning")
        timeout = self._cfg.getInt(Config.DEFAULT_TIMEOUT_KEY, 15)
        ntf.set_timeout(timeout * 1000)
        ntf.timeout = timeout * 1000
        if self.DOES_SUPPORT_ACTIONS:
            ntf.set_urgency(self.ntf2.URGENCY_CRITICAL)
            ntf.add_action(self.ACTION_ID_ALLOW, self.ACTION_ALLOW, callback, connection)
            ntf.add_action(self.ACTION_ID_DENY, self.ACTION_DENY, callback, connection)
            #ntf.add_action("open-gui", QC.translate("popups", "View"), callback, connection)
        ntf.set_category(self.CATEGORY_NETWORK)
        ntf.set_hint(self.HINT_DESKTOP_ENTRY, "opensnitch_ui")
        ntf.show()




from PyQt5 import QtCore
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtSql import QSqlQuery, QSqlError
from PyQt5.QtWidgets import QTableView
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QCoreApplication as QC

from opensnitch.nodes import Nodes
from opensnitch.firewall import Firewall
from opensnitch.customwidgets.updownbtndelegate import UpDownButtonDelegate

class FirewallTableModel(QStandardItemModel):
    rowCountChanged = pyqtSignal()
    columnCountChanged = pyqtSignal(int)
    rowsUpdated = pyqtSignal(int, tuple)
    rowsReordered = pyqtSignal(int, str, str, int, int) # filter, addr, key, old_pos, new_pos

    tableName = ""
    # total row count which must de displayed in the view
    totalRowCount = 0
    # last column count to compare against with
    lastColumnCount = 0

    FILTER_ALL = 0
    FILTER_BY_NODE = 1
    FILTER_BY_TABLE = 2
    FILTER_BY_CHAIN = 3
    FILTER_BY_QUERY = 4
    activeFilter = FILTER_ALL

    UP_BTN = -1
    DOWN_BTN = 1

    COL_BTNS = 0
    COL_UUID = 1
    COL_ADDR = 2
    COL_CHAIN_NAME = 3
    COL_CHAIN_TABLE = 4
    COL_CHAIN_FAMILY = 5
    COL_CHAIN_HOOK = 6
    COL_ENABLED = 7
    COL_DESCRIPTION = 8
    COL_PARMS = 9
    COL_ACTION = 10
    COL_ACTION_PARMS = 11

    headersAll = [
        "", # buttons
        "", # uuid
        QC.translate("firewall", "Node", ""),
        QC.translate("firewall", "Name", ""),
        QC.translate("firewall", "Table", ""),
        QC.translate("firewall", "Family", ""),
        QC.translate("firewall", "Hook", ""),
        QC.translate("firewall", "Enabled", ""),
        QC.translate("firewall", "Description", ""),
        QC.translate("firewall", "Parameters", ""),
        QC.translate("firewall", "Action", ""),
        QC.translate("firewall", "ActionParms", ""),
    ]

    items = []
    lastRules = []
    position = 0

    def __init__(self, tableName):
        self.tableName = tableName
        self._nodes = Nodes.instance()
        self._fw = Firewall.instance()
        self.lastColumnCount = len(self.headersAll)
        self.lastQueryArgs = ()

        QStandardItemModel.__init__(self, 0, self.lastColumnCount)
        self.setHorizontalHeaderLabels(self.headersAll)

    def filterByNode(self, addr):
        self.activeFilter = self.FILTER_BY_NODE
        self.fillVisibleRows(0, True, addr)

    def filterAll(self):
        self.activeFilter = self.FILTER_ALL
        self.fillVisibleRows(0, True)

    def filterByTable(self, addr, name, family):
        self.activeFilter = self.FILTER_BY_TABLE
        self.fillVisibleRows(0, True, addr, name, family)

    def filterByChain(self, addr, table, family, chain, hook):
        self.activeFilter = self.FILTER_BY_CHAIN
        self.fillVisibleRows(0, True, addr, table, family, chain, hook)

    def filterByQuery(self, query):
        self.activeFilter = self.FILTER_BY_QUERY
        self.fillVisibleRows(0, True, query)

    def reorderRows(self, action, row):
        if (row.row()+action == self.rowCount() and action == self.DOWN_BTN) or \
                (row.row() == 0 and action == self.UP_BTN):
            return

        # XXX: better use moveRow()?
        newRow = []
        # save the row we're about to overwrite
        for c in range(self.columnCount()):
            item = self.index(row.row()+action, c)
            itemText = item.data()
            newRow.append(itemText)
        # overwrite next item with current data
        for c in range(self.columnCount()):
            curItem = self.index(row.row(), c).data()
            nextIdx = self.index(row.row()+action, c)
            self.setData(nextIdx, curItem, QtCore.Qt.DisplayRole)
        # restore row with the overwritten data
        for i, nr in enumerate(newRow):
            idx = self.index(row.row(), i)
            self.setData(idx, nr, QtCore.Qt.DisplayRole)

        self.rowsReordered.emit(
            self.activeFilter,
            self.index(row.row()+action, self.COL_ADDR).data(), # address
            self.index(row.row()+action, self.COL_UUID).data(), # key
            row.row(),
            row.row()+action)

    def refresh(self, force=False):
        self.fillVisibleRows(0, force, *self.lastQueryArgs)

    #Some QSqlQueryModel methods must be mimiced so that this class can serve as a drop-in replacement
    #mimic QSqlQueryModel.query()
    def query(self):
        return self

    #mimic QSqlQueryModel.query().lastError()
    def lastError(self):
        return QSqlError()

    #mimic QSqlQueryModel.clear()
    def clear(self):
        self.items = []
        self.removeColumns(0, self.lastColumnCount)
        self.setColumnCount(0)
        self.setRowCount(0)

    # set columns based on query's fields
    def setModelColumns(self, headers):
        count = len(headers)
        self.clear()
        self.setHorizontalHeaderLabels(headers)
        self.lastColumnCount = count
        self.setColumnCount(self.lastColumnCount)
        self.columnCountChanged.emit(count)

    def query(self):
        return QSqlQuery()

    def setQuery(self, q, db, args=None):
        self.refresh()

    def nextRecord(self, offset):
        self.position += 1

    def prevRecord(self, offset):
        self.position -= 1

    def fillVisibleRows(self, upperBound, force, *data):
        if self.activeFilter == self.FILTER_BY_NODE and len(data) == 0:
                return

        cols = []
        rules = []
        #don't trigger setItem's signals for each cell, instead emit dataChanged for all cells
        self.blockSignals(True)
        # mandatory for rows refreshing
        self.layoutAboutToBeChanged.emit()

        if self.activeFilter == self.FILTER_BY_NODE:
            rules = self._fw.get_node_rules(data[0])
            self.setModelColumns(self.headersAll)
        elif self.activeFilter == self.FILTER_BY_TABLE:
            rules = self._fw.filter_by_table(data[0], data[1], data[2])
            self.setModelColumns(self.headersAll)
        elif self.activeFilter == self.FILTER_BY_CHAIN:
            rules = self._fw.filter_by_chain(data[0], data[1], data[2], data[3], data[4])
            self.setModelColumns(self.headersAll)
        elif self.activeFilter == self.FILTER_BY_QUERY:
            rules = self._fw.filter_rules(data[0])
            self.setModelColumns(self.headersAll)
        else:
            self.setModelColumns(self.headersAll)
            rules = self._fw.get_rules()

        self.addRows(rules)

        self.blockSignals(False)
        if self.lastRules != rules or force == True:
            self.layoutChanged.emit()
            self.totalRowCount = len(rules)
            self.setRowCount(self.totalRowCount)
            self.rowsUpdated.emit(self.activeFilter, data)
            self.dataChanged.emit(self.createIndex(0,0), self.createIndex(self.rowCount(), self.columnCount()))

        self.lastRules = rules
        self.lastQueryArgs = data
        del cols
        del rules

    def addRows(self, rules):
        self.items = []
        for rows in rules:
            cols = []
            cols.append(QStandardItem("")) # buttons column
            for cl in rows:
                item = QStandardItem(cl)
                item.setData(cl, QtCore.Qt.UserRole+1)
                cols.append(item)
            self.appendRow(cols)

    def dumpRows(self):
        for rule in self.lastRules:
            print(rule)

class FirewallTableView(QTableView):
    # how many rows can potentially be displayed in viewport
    # the actual number of rows currently displayed may be less than this
    maxRowsInViewport = 0
    rowsReordered = pyqtSignal(str) # addr

    def __init__(self, parent):
        QTableView.__init__(self, parent)
        self._fw = Firewall.instance()
        self._fw.rules.rulesUpdated.connect(self._cb_fw_rules_updated)

        self.verticalHeader().setVisible(True)
        self.horizontalHeader().setDefaultAlignment(QtCore.Qt.AlignCenter)
        self.horizontalHeader().setStretchLastSection(True)

        # FIXME: if the firewall being used is iptables, hide the column to
        # reorder rules, it's not supported.
        updownBtn = UpDownButtonDelegate(self)
        self.setItemDelegateForColumn(0, updownBtn)
        updownBtn.clicked.connect(self._cb_fw_rule_position_changed)

    def _cb_fw_rules_updated(self):
        self.model().refresh(True)

    def _cb_column_count_changed(self, num):
        for i in range(num):
            self.resizeColumnToContents(i)

    def _cb_fw_rule_position_changed(self, action, row):
        self.model().reorderRows(action, row)

    def _cb_rows_reordered(self, view, node_addr, uuid, old_pos, new_pos):
        if self._fw.swap_rules(view, node_addr, uuid, old_pos, new_pos):
            self.rowsReordered.emit(node_addr)

    #@QtCore.pyqtSlot(int, tuple)
    def _cb_rows_updated(self, view, data):
        for c in range(self.model().rowCount()):
            self.setColumnHidden(c, False)
            #self.horizontalHeader().setSectionResizeMode(
            #    c, QHeaderView.ResizeToContents
            #)

        self.setColumnHidden(FirewallTableModel.COL_BTNS, True)
        self.setColumnHidden(FirewallTableModel.COL_UUID, True)
        if view >= self.model().FILTER_BY_NODE:
            # hide address column
            self.setColumnHidden(FirewallTableModel.COL_ADDR, True)
        if view >= self.model().FILTER_BY_TABLE:
            self.setColumnHidden(FirewallTableModel.COL_CHAIN_TABLE, True)
            self.setColumnHidden(FirewallTableModel.COL_CHAIN_FAMILY, True)
        if view >= self.model().FILTER_BY_CHAIN:
            # hide chain's name, family and hook
            self.setColumnHidden(FirewallTableModel.COL_CHAIN_NAME, True)
            self.setColumnHidden(FirewallTableModel.COL_CHAIN_HOOK, True)
            self.setColumnHidden(FirewallTableModel.COL_BTNS, False)

    def filterAll(self):
        self.model().filterAll()

    def filterByNode(self, addr):
        self.model().filterByNode(addr)

    def filterByTable(self, addr, name, family):
        self.model().filterByTable(addr, name, family)

    def filterByChain(self, addr, table, family, chain, hook):
        self.model().filterByChain(addr, table, family, chain, hook)

    def filterByQuery(self, query):
        self.model().filterByQuery(query)

    def refresh(self):
        self.model().refresh(True)

    def clearSelection(self):
        pass

    def selectedRows(self):
        selection = self.selectedIndexes()
        if not selection:
            return None
        rows = []
        row = []
        lastRow = 0
        for idx in selection:
            if idx.row() == lastRow:
                row.append(self.model().index(idx.row(), idx.column()).data())
            else:
                row = []
                lastRow = idx.row()
                rows.append(row)
        return rows

    def setModel(self, model):
        super().setModel(model)
        self.horizontalHeader().sortIndicatorChanged.disconnect()
        self.setSortingEnabled(True)
        self.model().columnCountChanged.connect(self._cb_column_count_changed)
        model.rowsUpdated.connect(self._cb_rows_updated)
        model.rowsReordered.connect(self._cb_rows_reordered)

    def setTrackingColumn(self, col):
        pass



from PyQt5.QtSql import QSqlDatabase, QSqlQueryModel, QSqlQuery
import threading
import sys
import os
from datetime import datetime, timedelta

class Database:
    db = None
    __instance = None
    DB_IN_MEMORY   = "file::memory:"
    DB_TYPE_MEMORY = 0
    DB_TYPE_FILE   = 1
    DB_JRNL_WAL    = False

    # Sqlite3 journal modes
    DB_JOURNAL_MODE_LIST = {
            0: "DELETE",
            1: "TRUNCATE",
            2: "PERSIST",
            3: "MEMORY",
            4: "WAL",
            5: "OFF",
            }

    # increase accordingly whenever the schema is updated
    DB_VERSION = 3

    @staticmethod
    def instance():
        if Database.__instance == None:
            Database.__instance = Database()
        return Database.__instance

    def __init__(self, dbname="db"):
        self._lock = threading.RLock()
        self.db = None
        self.db_file = Database.DB_IN_MEMORY
        self.db_jrnl_wal = Database.DB_JRNL_WAL
        self.db_name = dbname

    def initialize(self, dbtype=DB_TYPE_MEMORY, dbfile=DB_IN_MEMORY, dbjrnl_wal=DB_JRNL_WAL, db_name="db"):
        if dbtype != Database.DB_TYPE_MEMORY:
            self.db_file = dbfile
            self.db_jrnl_wal = dbjrnl_wal
        else:
            # Always disable under pure memory mode
            self.db_jrnl_wal = False

        is_new_file = not os.path.isfile(self.db_file)

        self.db = QSqlDatabase.addDatabase("QSQLITE", self.db_name)
        self.db.setDatabaseName(self.db_file)
        if dbtype == Database.DB_TYPE_MEMORY:
            self.db.setConnectOptions("QSQLITE_OPEN_URI;QSQLITE_ENABLE_SHARED_CACHE")
        if not self.db.open():
            print("\n ** Error opening DB: SQLite driver not loaded. DB name: %s\n" % self.db_file)
            print("\n    Available drivers: ", QSqlDatabase.drivers())
            sys.exit(-1)

        db_status, db_error = self.is_db_ok()
        if db_status is False:
            print("db.initialize() error:", db_error)
            return False, db_error


        if is_new_file:
            print("is new file, or IN MEMORY, setting initial schema version")
            self.set_schema_version(self.DB_VERSION)

        self._create_tables()
        self._upgrade_db_schema()
        return True, None

    def close(self):
        try:
            if self.db.isOpen():
                self.db.removeDatabase(self.db_name)
                self.db.close()
        except Exception as e:
            print("db.close() exception:", e)

    def is_db_ok(self):
        # XXX: quick_check may not be fast enough with some DBs on slow
        # hardware.
        q = QSqlQuery("PRAGMA quick_check;", self.db)
        if q.exec_() is not True:
            print(q.lastError().driverText())
            return False, q.lastError().driverText()

        if q.next() and q.value(0) != "ok":
            return False, "Database is corrupted (1)"

        return True, None

    def get_db(self):
        return self.db

    def get_db_file(self):
        return self.db_file

    def get_new_qsql_model(self):
        return QSqlQueryModel()

    def get_db_name(self):
        return self.db_name

    def _create_tables(self):
        # https://www.sqlite.org/wal.html
        if self.db_file == Database.DB_IN_MEMORY:
            self.set_schema_version(self.DB_VERSION)
            # Disable journal (default)
            self.set_journal_mode(5)
        elif self.db_jrnl_wal is True:
            # Set WAL mode (file+memory)
            self.set_journal_mode(4)
        else:
            # Set DELETE mode (file)
            self.set_journal_mode(0)
        q = QSqlQuery("create table if not exists connections (" \
                "time text, " \
                "node text, " \
                "action text, " \
                "protocol text, " \
                "src_ip text, " \
                "src_port text, " \
                "dst_ip text, " \
                "dst_host text, " \
                "dst_port text, " \
                "uid text, " \
                "pid text, " \
                "process text, " \
                "process_args text, " \
                "process_cwd text, " \
                "rule text, " \
                "UNIQUE(node, action, protocol, src_ip, src_port, dst_ip, dst_port, uid, pid, process, process_args))",
                self.db)
        q = QSqlQuery("create index time_index on connections (time)", self.db)
        q.exec_()
        q = QSqlQuery("create index action_index on connections (action)", self.db)
        q.exec_()
        q = QSqlQuery("create index protocol_index on connections (protocol)", self.db)
        q.exec_()
        q = QSqlQuery("create index dst_host_index on connections (dst_host)", self.db)
        q.exec_()
        q = QSqlQuery("create index process_index on connections (process)", self.db)
        q.exec_()
        q = QSqlQuery("create index dst_ip_index on connections (dst_ip)", self.db)
        q.exec_()
        q = QSqlQuery("create index dst_port_index on connections (dst_port)", self.db)
        q.exec_()
        q = QSqlQuery("create index rule_index on connections (rule)", self.db)
        q.exec_()
        q = QSqlQuery("create index node_index on connections (node)", self.db)
        q.exec_()
        q = QSqlQuery("CREATE INDEX details_query_index on connections (process, process_args, uid, pid, dst_ip, dst_host, dst_port, action, node, protocol)", self.db)
        q.exec_()

        q = QSqlQuery("create table if not exists nodes (" \
                "addr text primary key," \
                "hostname text," \
                "daemon_version text," \
                "daemon_uptime text," \
                "daemon_rules text," \
                "cons text," \
                "cons_dropped text," \
                "version text," \
                "status text, " \
                "last_connection text)"
                , self.db)
        q.exec_()

        q = QSqlQuery("create table if not exists rules (" \
                "time text, " \
                "node text, " \
                "name text, " \
                "enabled text, " \
                "precedence text, " \
                "action text, " \
                "duration text, " \
                "operator_type text, " \
                "operator_sensitive text, " \
                "operator_operand text, " \
                "operator_data text, " \
                "description text, " \
                "nolog text, " \
                "created text, " \
                "UNIQUE(node, name)"
                ")", self.db)
        q.exec_()

        q = QSqlQuery("create table if not exists alerts (" \
                "time text, " \
                "node text, " \
                "type text, " \
                "action text, " \
                "priority text, " \
                "what text, " \
                "body text, " \
                "status int " \
                ")", self.db)
        q.exec_()
        q = QSqlQuery("create table if not exists sockets (" \
                      "id int primary key, " \
                      "last_seen text, " \
                      "node text, " \
                      "src_port text, " \
                      "src_ip text, " \
                      "dst_ip text, " \
                      "dst_port text, " \
                      "proto text, " \
                      "uid text, " \
                      "inode text, " \
                      "iface text, " \
                      "family text, " \
                      "state text, " \
                      "cookies text, " \
                      "rqueue text, " \
                      "wqueue text, " \
                      "expires text, " \
                      "retrans text, " \
                      "timer text, " \
                      "mark text, " \
                      "proc_pid text, " \
                      "proc_comm text, " \
                      "proc_path text, " \
                      "UNIQUE(node, src_port, src_ip, dst_ip, dst_port, proto, family, inode)" \
                      ")", self.db)
        q.exec_()
        q = QSqlQuery("create index sck_srcport_index on sockets (src_port)", self.db)
        q.exec_()
        q = QSqlQuery("create index sck_dstip_index on sockets (dst_ip)", self.db)
        q.exec_()
        q = QSqlQuery("create index sck_srcip_index on sockets (src_ip)", self.db)
        q.exec_()
        q = QSqlQuery("create index sck_dsthost_index on sockets (dst_host)", self.db)
        q.exec_()
        q = QSqlQuery("create index sck_state_index on sockets (state)", self.db)
        q.exec_()
        q = QSqlQuery("create index sck_comm_index on sockets (proc_comm)", self.db)
        q.exec_()
        q = QSqlQuery("create index sck_path_index on sockets (proc_path)", self.db)
        q.exec_()
        q.exec_()
        q = QSqlQuery("create index rules_index on rules (time)", self.db)
        q.exec_()

        q = QSqlQuery("create table if not exists hosts (what text primary key, hits integer)", self.db)
        q.exec_()
        q = QSqlQuery("create table if not exists procs (what text primary key, hits integer)", self.db)
        q.exec_()
        q = QSqlQuery("create table if not exists addrs (what text primary key, hits integer)", self.db)
        q.exec_()
        q = QSqlQuery("create table if not exists ports (what text primary key, hits integer)", self.db)
        q.exec_()
        q = QSqlQuery("create table if not exists users (what text primary key, hits integer)", self.db)
        q.exec_()

    def get_schema_version(self):
        q = QSqlQuery("PRAGMA user_version;", self.db)
        q.exec_()
        if q.next():
            print("schema version:", q.value(0))
            return int(q.value(0))

        return 0

    def set_schema_version(self, version):
        print("setting schema version to:", version)
        q = QSqlQuery("PRAGMA user_version = {0}".format(version), self.db)
        if q.exec_() == False:
            print("Error updating updating schema version:", q.lastError().text())

    def get_journal_mode(self):
        q = QSqlQuery("PRAGMA journal_mode;", self.db)
        q.exec_()
        if q.next():
            return str(q.value(0))

        return str("unknown")

    def set_journal_mode(self, mode):
        # https://www.sqlite.org/wal.html
        mode_str = Database.DB_JOURNAL_MODE_LIST[mode]
        if self.get_journal_mode().lower() != mode_str.lower():
            print("Setting journal_mode: ", mode_str)
            q = QSqlQuery("PRAGMA journal_mode = {modestr};".format(modestr = mode_str), self.db)
            if q.exec_() == False:
                print("Error updating PRAGMA journal_mode:", q.lastError().text())
                return False
        if mode == 3 or mode == 5:
            print("Setting DB memory optimizations")
            q = QSqlQuery("PRAGMA synchronous = OFF;", self.db)
            if q.exec_() == False:
                print("Error updating PRAGMA synchronous:", q.lastError().text())
                return False
            q = QSqlQuery("PRAGMA cache_size=10000;", self.db)
            if q.exec_() == False:
                print("Error updating PRAGMA cache_size:", q.lastError().text())
                return False
        else:
            print("Setting synchronous = NORMAL")
            q = QSqlQuery("PRAGMA synchronous = NORMAL;", self.db)
            if q.exec_() == False:
                print("Error updating PRAGMA synchronous:", q.lastError().text())

        return True

    def _upgrade_db_schema(self):
        migrations_path = os.path.dirname(os.path.realpath(__file__)) + "/migrations"
        schema_version = self.get_schema_version()
        if schema_version == self.DB_VERSION:
            print("db schema is up to date")
            return
        while schema_version < self.DB_VERSION:
            schema_version += 1
            try:
                print("applying schema upgrade:", schema_version)
                self._apply_db_upgrade("{0}/upgrade_{1}.sql".format(migrations_path, schema_version))
            except Exception as e:
                print("Not applying upgrade_{0}.sql:".format(schema_version), e)
                return
        self.set_schema_version(schema_version)

    def _apply_db_upgrade(self, file):
        print("applying upgrade from:", file)
        q = QSqlQuery(self.db)
        with open(file) as f:
            for line in f.readlines():
                # skip comments
                if line.startswith("--"):
                    continue
                print("applying upgrade:", line, end="")
                if q.exec(line) == False:
                    print("\tError:", q.lastError().text())
                else:
                    print("\tOK")

    def optimize(self):
        """https://www.sqlite.org/pragma.html#pragma_optimize
        """
        q = QSqlQuery("PRAGMA optimize;", self.db)
        q.exec_()

    def clean(self, table):
        with self._lock:
            q = QSqlQuery("delete from " + table, self.db)
            q.exec_()

    def vacuum(self):
        q = QSqlQuery("VACUUM;", self.db)
        q.exec_()

    def clone_db(self, name):
        return QSqlDatabase.cloneDatabase(self.db, name)

    def clone(self):
        q = QSqlQuery(".dump", self.db)
        q.exec_()

    def transaction(self):
        self.db.transaction()

    def commit(self):
        self.db.commit()

    def rollback(self):
        self.db.rollback()

    def get_total_records(self):
        try:
            q = QSqlQuery("SELECT count(*) FROM connections", self.db)
            if q.exec_() and q.first():
                r = q.value(0)
        except Exception as e:
            print("db, get_total_records() error:", e)

    def get_newest_record(self):
        try:
            q = QSqlQuery("SELECT time FROM connections ORDER BY 1 DESC LIMIT 1", self.db)
            if q.exec_() and q.first():
                return q.value(0)
        except Exception as e:
            print("db, get_newest_record() error:", e)
        return 0

    def get_oldest_record(self):
        try:
            q = QSqlQuery("SELECT time FROM connections ORDER BY 1 ASC LIMIT 1", self.db)
            if q.exec_() and q.first():
                return q.value(0)
        except Exception as e:
            print("db, get_oldest_record() error:", e)
        return 0

    def purge_oldest(self, max_days_to_keep):
        try:
            oldt = self.get_oldest_record()
            newt = self.get_newest_record()
            if oldt == None or newt == None or oldt == 0 or newt == 0:
                return -1

            oldest = datetime.strptime(oldt, "%Y-%m-%d %H:%M:%S.%f")
            newest = datetime.strptime(newt, "%Y-%m-%d %H:%M:%S.%f")
            diff = newest - oldest
            date_to_purge = datetime.now() - timedelta(days=max_days_to_keep)

            if diff.days >= max_days_to_keep:
                q = QSqlQuery(self.db)
                q.prepare("DELETE FROM connections WHERE time < ?")
                q.bindValue(0, str(date_to_purge))
                if q.exec_():
                    print("purge_oldest() {0} records deleted".format(q.numRowsAffected()))
                    return q.numRowsAffected()
        except Exception as e:
            print("db, purge_oldest() error:", e)

        return -1

    def select(self, qstr):
        try:
            return QSqlQuery(qstr, self.db)
        except Exception as e:
            print("db, select() exception: ", e)

        return None

    def remove(self, qstr):
        try:
            q = QSqlQuery(qstr, self.db)
            if q.exec_():
                return True
            else:
                print("db, remove() ERROR: ", qstr)
                print(q.lastError().driverText())
        except Exception as e:
            print("db, remove exception: ", e)

        return False

    def _insert(self, query_str, columns):
        with self._lock:
            try:

                q = QSqlQuery(self.db)
                q.prepare(query_str)
                for idx, v in enumerate(columns):
                    q.bindValue(idx, v)
                if q.exec_():
                    return True
                else:
                    print("_insert() ERROR", query_str)
                    print(q.lastError().driverText())

            except Exception as e:
                print("_insert exception", e)
            finally:
                q.finish()

        return False

    def insert(self, table, fields, columns, update_field=None, update_values=None, action_on_conflict="REPLACE"):
        if update_field != None:
            action_on_conflict = ""
        else:
            action_on_conflict = "OR " + action_on_conflict

        qstr = "INSERT " + action_on_conflict + " INTO " + table + " " + fields + " VALUES("
        update_fields=""
        for col in columns:
            qstr += "?,"
        qstr = qstr[0:len(qstr)-1] + ")"

        if update_field != None:
            # NOTE: UPSERTS on sqlite are only supported from v3.24 on.
            # On Ubuntu16.04/18 for example (v3.11/3.22) updating a record on conflict
            # fails with "Parameter count error"
            qstr += " ON CONFLICT (" + update_field + ") DO UPDATE SET "
            for idx, field in enumerate(update_values):
                qstr += str(field) + "=excluded." + str(field) + ","

            qstr = qstr[0:len(qstr)-1]

        return self._insert(qstr, columns)

    def update(self, table, fields, values, condition=None, action_on_conflict="OR IGNORE"):
        qstr = "UPDATE " + action_on_conflict + " " + table + " SET " + fields
        if condition != None:
            qstr += " WHERE " + condition
        try:
            with self._lock:
                q = QSqlQuery(qstr, self.db)
                q.prepare(qstr)
                for idx, v in enumerate(values):
                    q.bindValue(idx, v)
                if not q.exec_():
                    print("update ERROR:", qstr, "values:", values)
                    print(q.lastError().driverText())

        except Exception as e:
            print("update() exception:", e)
        finally:
            q.finish()

    def _insert_batch(self, query_str, fields, values):
        result=True
        with self._lock:
            try:
                q = QSqlQuery(self.db)
                q.prepare(query_str)
                q.addBindValue(fields)
                q.addBindValue(values)
                if not q.execBatch():
                    print("_insert_batch() db error:", query_str)
                    print(q.lastError().driverText())
                    print("\t", fields)
                    print("\t", values)

                    result=False
            except Exception as e:
                print("_insert_batch() exception:", e)
            finally:
                q.finish()

        return result

    def insert_batch(self, table, db_fields, db_columns, fields, values, update_field=None, update_value=None, action_on_conflict="REPLACE"):
        action = "OR " + action_on_conflict
        if update_field != None:
            action = ""

        qstr = "INSERT " + action + " INTO " + table + " (" + db_fields[0] + "," + db_fields[1] + ") VALUES("
        for idx in db_columns:
            qstr += "?,"
        qstr = qstr[0:len(qstr)-1] + ")"

        if self._insert_batch(qstr, fields, values) == False:
            self.update_batch(table, db_fields, db_columns, fields, values, update_field, update_value, action_on_conflict)

    def update_batch(self, table, db_fields, db_columns, fields, values, update_field=None, update_value=None, action_on_conflict="REPLACE"):
        for idx, i in enumerate(values):
            s = "UPDATE " + table + " SET " + "%s=(select hits from %s)+%s" % (db_fields[1], table, values[idx])
            s += "  WHERE %s=\"%s\"," % (db_fields[0], fields[idx])
            s = s[0:len(s)-1]
            with self._lock:
                q = QSqlQuery(s, self.db)
                if not q.exec_():
                    print("update batch ERROR", s)
                    print(q.lastError().driverText())

    def dump(self):
        q = QSqlQuery(".dump", db=self.db)
        q.exec_()

    def get_query(self, table, fields):
        return "SELECT " + fields + " FROM " + table

    def empty_rule(self, name=""):
        if name == "":
            return
        qstr = "DELETE FROM connections WHERE rule = ?"

        with self._lock:
            q = QSqlQuery(qstr, self.db)
            q.prepare(qstr)
            q.addBindValue(name)
            if not q.exec_():
                print("db, empty_rule() ERROR: ", qstr)
                print(q.lastError().driverText())

    def delete_rule(self, name, node_addr):
        qstr = "DELETE FROM rules WHERE name=?"
        if node_addr != None:
            qstr = qstr + " AND node=?"

        with self._lock:
            q = QSqlQuery(qstr, self.db)
            q.prepare(qstr)
            q.addBindValue(name)
            if node_addr != None:
                q.addBindValue(node_addr)
            if not q.exec_():
                print("db, delete_rule() ERROR: ", qstr)
                print(q.lastError().driverText())
                return False

        return True

    def delete_rules_by_field(self, field, values):
        if len(values) == 0:
            return True

        qstr = "DELETE FROM rules WHERE "
        for v in values:
            qstr += field + "=? OR "

        qstr = qstr[:-4]

        with self._lock:
            q = QSqlQuery(qstr, self.db)
            q.prepare(qstr)

            for v in values:
                q.addBindValue(v)

            if not q.exec_():
                print("db, delete_rule_by_field() ERROR: ", qstr)
                print(q.lastError().driverText())
                return False

        return True

    def get_connection_by_field(self, field, date):
        """
        """
        qstr = "SELECT * FROM connections WHERE {0}=?".format(field)

        q = QSqlQuery(qstr, self.db)
        q.prepare(qstr)
        q.addBindValue(date)
        q.exec_()

        return q

    def get_rule(self, rule_name, node_addr=None):
        """
        get rule records, given the name of the rule and the node
        """
        qstr = "SELECT * FROM rules WHERE name=?"
        if node_addr != None:
            qstr = qstr + " AND node=?"

        q = QSqlQuery(qstr, self.db)
        q.prepare(qstr)
        q.addBindValue(rule_name)
        if node_addr != None:
            q.addBindValue(node_addr)
        q.exec_()

        return q

    def get_rule_by_field(self, node_addr=None, field=None, value=None):
        """
        get rule records by field (process.path, etc)
        """
        qstr = "SELECT * FROM rules WHERE {0} LIKE ?".format(field)
        q = QSqlQuery(qstr, self.db)
        if node_addr != None:
            qstr = qstr + " AND node=?".format(node_addr)

        q.prepare(qstr)
        q.addBindValue("%" + value + "%")
        if node_addr != None:
            q.addBindValue(node_addr)
        if not q.exec_():
            print("get_rule_by_field() error:", q.lastError().driverText())
            return None

        return q

    def get_rules(self, node_addr):
        """
        get rule records, given the name of the rule and the node
        """
        qstr = "SELECT * FROM rules WHERE node=?"
        q = QSqlQuery(qstr, self.db)
        q.prepare(qstr)
        q.addBindValue(node_addr)
        if not q.exec_():
            return None

        return q

    def insert_rule(self, rule, node_addr):
        self.insert("rules",
            "(time, node, name, description, enabled, precedence, nolog, action, duration, operator_type, operator_sensitive, operator_operand, operator_data)",
                (datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                    node_addr, rule.name, rule.description,
                    str(rule.enabled), str(rule.precedence), str(rule.nolog),
                    rule.action, rule.duration, rule.operator.type,
                    str(rule.operator.sensitive), rule.operator.operand, rule.operator.data),
                action_on_conflict="IGNORE")

    def rule_exists(self, rule, node_addr):
        qstr = "SELECT node, name, action, duration, operator_type, operator_operand, operator_data " \
            " FROM rules WHERE " \
            "name=? AND " \
            "node=? AND " \
            "action=? AND " \
            "duration=? AND " \
            "operator_type=? AND " \
            "operator_operand=? AND " \
            "operator_data=?"
        q = QSqlQuery(qstr, self.db)
        q.prepare(qstr)
        q.addBindValue(rule.name)
        q.addBindValue(node_addr)
        q.addBindValue(rule.action)
        q.addBindValue(rule.duration)
        q.addBindValue(rule.operator.type)
        q.addBindValue(rule.operator.operand)
        q.addBindValue(rule.operator.data)
        if not q.exec_() or q.next() == False:
            return None

        return q


    def delete_alert(self, time, node_addr=None):
        qstr = "DELETE FROM alerts WHERE time=?"
        if node_addr != None:
            qstr = qstr + " AND node=?"

        with self._lock:
            q = QSqlQuery(qstr, self.db)
            q.prepare(qstr)
            q.addBindValue(time)
            if node_addr != None:
                q.addBindValue(node_addr)
            if not q.exec_():
                print("db, delete_alert() ERROR: ", qstr)
                print(q.lastError().driverText())
                return False

        return True

    def get_alert(self, alert_time, node_addr=None):
        """
        get alert, given the time of the alert and the node
        """
        qstr = "SELECT * FROM alerts WHERE time=?"
        if node_addr != None:
            qstr = qstr + " AND node=?"

        q = QSqlQuery(qstr, self.db)
        q.prepare(qstr)
        q.addBindValue(alert_time)
        if node_addr != None:
            q.addBindValue(node_addr)
        q.exec_()

        return q




import glob
import json
import os.path


class Profiles():

    @staticmethod
    def load_predefined_profiles():
        profiles = glob.glob("/etc/opensnitchd/system-fw.d/profiles/*.profile")
        p = []
        for pr_path in profiles:
            with open(pr_path) as f:
                p.append({os.path.basename(pr_path): json.load(f)})

        return p


class ProfileAcceptOutput():
    value = {
        "Name": "accept-mangle-output",
        "Table": "mangle",
        "Family": "inet",
        "Priority": "",
        "Type": "mangle",
        "Hook": "output",
        "Policy": "accept",
        "Rules": [
        ]
    }

class ProfileAcceptForward():
    value = {
        "Name": "accept-mangle-forward",
        "Table": "mangle",
        "Family": "inet",
        "Priority": "",
        "Type": "mangle",
        "Hook": "forward",
        "Policy": "accept",
        "Rules": [
        ]
    }


class ProfileDropForward():
    value = {
        "Name": "drop-mangle-forward",
        "Table": "mangle",
        "Family": "inet",
        "Priority": "",
        "Type": "mangle",
        "Hook": "forward",
        "Policy": "drop",
        "Rules": [
        ]
    }


class ProfileAcceptInput():
    value = {
        "Name": "accept-filter-input",
        "Table": "filter",
        "Family": "inet",
        "Priority": "",
        "Type": "filter",
        "Hook": "input",
        "Policy": "accept",
        "Rules": [
        ]
    }


class ProfileDropInput():
    """
    Set input filter table policy to DROP and add the needed rules to allow
    outbound connections.
    """

    # TODO: delete dropInput profile's rules
    value = {
        "Name": "drop-filter-input",
        "Table": "filter",
        "Family": "inet",
        "Priority": "",
        "Type": "filter",
        "Hook": "input",
        "Policy": "drop",
        "Rules": [
            {
                "Table": "",
                "Chain": "",
                "UUID": "profile-drop-inbound-2d7e6fe4-c21d-11ec-99a6-3c970e298b0c",
                "Enabled": True,
                "Position": "0",
                "Description": "[profile-drop-inbound] allow localhost connections",
                "Parameters": "",
                "Expressions": [
                    {
                        "Statement": {
                            "Op": "",
                            "Name": "iifname",
                            "Values": [
                                {
                                    "Key": "lo",
                                    "Value": ""
                                }
                            ]
                        }
                    }
                ],
                "Target": "accept",
                "TargetParameters": ""
            },
            {
                "Enabled": True,
                "Description": "[profile-drop-inbound] allow established,related connections",
                "UUID": "profile-drop-inbound-e1fc1a1c-c21c-11ec-9a2a-3c970e298b0c",
                "Expressions": [
                    {
                        "Statement": {
                            "Op": "",
                            "Name": "ct",
                            "Values": [
                                {
                                    "Key": "state",
                                    "Value": "related"
                                },
                                {
                                    "Key": "state",
                                    "Value": "established"
                                }
                            ]
                        }
                    }
                ],
                "Target": "accept",
                "TargetParameters": ""
            }
        ]
    }

class ProfileDropOutput():
    """
    Set output mangle table policy to DROP and add the needed rules to allow
    outbound connections.
    """

    value = {
        "Name": "drop-mangle-output",
        "Table": "mangle",
        "Family": "inet",
        "Priority": "",
        "Type": "mangle",
        "Hook": "output",
        "Policy": "drop",
        "Rules": [
            {
                "Enabled": True,
                "Description": "[profile-drop-outbound] allow established,related connections",
                "UUID": "profile-drop-outbound-e1fc1a1c-c21c-11ec-9a2a-3c970e298b0c",
                "Expressions": [
                    {
                        "Statement": {
                            "Op": "",
                            "Name": "ct",
                            "Values": [
                                {
                                    "Key": "state",
                                    "Value": "established,related"
                                }
                            ]
                        }
                    }
                ],
                "Target": "accept",
                "TargetParameters": ""
            }
        ]
    }



from PyQt5.QtGui import QColor

def getColorNames():
    """Return the built-in color names that can be used to choose new colors:
        https://doc.qt.io/qtforpython-5/PySide2/QtGui/QColor.html#predefined-colors
        https://www.w3.org/TR/SVG11/types.html#ColorKeywords
    """
    return QColor.colorNames()



from PyQt5.QtGui import QStandardItemModel
from PyQt5.QtSql import QSqlQuery, QSql
from PyQt5.QtWidgets import QTableView
from PyQt5.QtCore import (
    QItemSelectionRange,
    QItemSelectionModel,
    QItemSelection,
    QModelIndex,
    pyqtSignal,
    QEvent,
    Qt)
import math

class GenericTableModel(QStandardItemModel):
    rowCountChanged = pyqtSignal()
    beginViewPortRefresh = pyqtSignal()
    endViewPortRefresh = pyqtSignal()

    db = None
    tableName = ""
    # total row count which must de displayed in the view
    totalRowCount = 0
    #
    lastColumnCount = 0

    # original query string before we modify it
    origQueryStr = QSqlQuery()
    # previous original query string; used to check if the query has changed
    prevQueryStr = ''
    # modified query object
    realQuery = QSqlQuery()

    items = []
    lastItems = []

    def __init__(self, tableName, headerLabels):
        self.tableName = tableName
        self.headerLabels = headerLabels
        self.lastColumnCount = len(self.headerLabels)
        QStandardItemModel.__init__(self, 0, self.lastColumnCount)
        self.setHorizontalHeaderLabels(self.headerLabels)

    #Some QSqlQueryModel methods must be mimiced so that this class can serve as a drop-in replacement
    #mimic QSqlQueryModel.query()
    def query(self):
        return self

    #mimic QSqlQueryModel.query().lastQuery()
    def lastQuery(self):
        return self.origQueryStr

    #mimic QSqlQueryModel.query().lastError()
    def lastError(self):
        return self.realQuery.lastError()

    #mimic QSqlQueryModel.clear()
    def clear(self):
        pass

    def rowCount(self, index=None):
        """ensures that only the needed rows is created"""
        return len(self.items)

    def data(self, index, role=Qt.DisplayRole):
        """Paint rows with the data stored in self.items
        """
        if role == Qt.DisplayRole or role == Qt.EditRole:
            items_count = len(self.items)
            if index.isValid() and items_count > 0 and index.row() < items_count:
                return self.items[index.row()][index.column()]
        return QStandardItemModel.data(self, index, role)

    # set columns based on query's fields
    def setModelColumns(self, newColumns):
        # Avoid firing signals while reconfiguring the view, it causes
        # segfaults.
        self.blockSignals(True);

        self.headerLabels = []
        self.removeColumns(0, self.lastColumnCount)
        self.setHorizontalHeaderLabels(self.headerLabels)
        for col in range(0, newColumns):
            self.headerLabels.append(self.realQuery.record().fieldName(col))
        self.lastColumnCount = newColumns
        self.setHorizontalHeaderLabels(self.headerLabels)
        self.setColumnCount(len(self.headerLabels))

        self.blockSignals(False);

    def setQuery(self, q, db):
        self.origQueryStr = q
        self.db = db
        #print("q:", q)

        if self.prevQueryStr != self.origQueryStr:
            self.realQuery = QSqlQuery(q, db)

        self.realQuery.exec_()
        self.realQuery.last()

        queryRows = max(0, self.realQuery.at()+1)
        self.totalRowCount = queryRows
        self.setRowCount(self.totalRowCount)

        # update view's columns
        queryColumns = self.realQuery.record().count()
        if queryColumns != self.lastColumnCount:
            self.setModelColumns(queryColumns)

        self.prevQueryStr = self.origQueryStr
        self.rowCountChanged.emit()

    def nextRecord(self, offset):
        cur_pos = self.realQuery.at()
        q.seek(max(cur_pos, cur_pos+offset))

    def prevRecord(self, offset):
        cur_pos = self.realQuery.at()
        q.seek(min(cur_pos, cur_pos-offset))

    def refreshViewport(self, scrollValue, maxRowsInViewport, force=False):
        """Refresh the viewport with data from the db.
        Before making any changes, emit a signal which will perform several operations
        (save current selected row, etc).
        force var will force a refresh if the scrollbar is at the top or bottom of the
        viewport, otherwise skip it to allow rows analyzing without refreshing.
        """
        if not force:
            return

        self.beginViewPortRefresh.emit()
        # set records position to last, in order to get correctly the number of
        # rows.
        self.realQuery.last()
        rowsFound = max(0, self.realQuery.at()+1)
        if scrollValue == 0 or self.realQuery.at() == QSql.BeforeFirstRow:
            self.realQuery.seek(QSql.BeforeFirstRow)
        elif self.realQuery.at() == QSql.AfterLastRow:
            self.realQuery.seek(rowsFound - maxRowsInViewport)
        else:
            self.realQuery.seek(min(scrollValue-1, self.realQuery.at()))

        upperBound = min(maxRowsInViewport, rowsFound)
        self.setRowCount(self.totalRowCount)

        # only visible rows will be filled with data, and only if we're not
        # updating the viewport already.
        if force and (upperBound > 0 or self.realQuery.at() < 0):
            self.fillVisibleRows(self.realQuery, upperBound, force)
        self.endViewPortRefresh.emit()

    def fillVisibleRows(self, q, upperBound, force=False):
        rowsLabels = []
        self.setVerticalHeaderLabels(rowsLabels)

        self.items = []
        cols = []
        #don't trigger setItem's signals for each cell, instead emit dataChanged for all cells
        for x in range(0, upperBound):
            q.next()
            if q.at() < 0:
                # if we don't set query to a valid record here, it gets stucked
                # forever at -2/-1.
                q.seek(upperBound)
                break
            rowsLabels.append(str(q.at()+1))
            cols = []
            for col in range(0, len(self.headerLabels)):
                cols.append(str(q.value(col)))

            self.items.append(cols)

        self.setVerticalHeaderLabels(rowsLabels)
        if self.lastItems != self.items or force == True:
            self.dataChanged.emit(self.createIndex(0,0), self.createIndex(upperBound, len(self.headerLabels)))
        self.lastItems = self.items
        del cols

    def dumpRows(self):
        rows = []
        q = QSqlQuery(self.db)
        q.exec(self.origQueryStr)
        q.seek(QSql.BeforeFirstRow)
        while True:
            q.next()
            if q.at() == QSql.AfterLastRow:
                break
            row = []
            for col in range(0, len(self.headerLabels)):
                row.append(q.value(col))
            rows.append(row)
        return rows

    def copySelectedRows(self, start=QSql.BeforeFirstRow, end=QSql.AfterLastRow):
        rows = []
        lastAt = self.realQuery.at()
        self.realQuery.seek(start)
        while True:
            self.realQuery.next()
            if self.realQuery.at() == QSql.AfterLastRow or len(rows) >= end:
                break
            row = []
            for col in range(0, len(self.headerLabels)):
                row.append(self.realQuery.value(col))
            rows.append(row)
        self.realQuery.seek(lastAt)
        return rows

class GenericTableView(QTableView):
    # how many rows can potentially be displayed in viewport
    # the actual number of rows currently displayed may be less than this
    maxRowsInViewport = 0
    vScrollBar = None

    def __init__(self, parent):
        QTableView.__init__(self, parent)
        self.mousePressed = False
        self.shiftPressed = False
        self.ctrlPressed = False
        self._rows_selection = {}
        self.trackingCol = 0

        #eventFilter to catch key up/down events and wheel events
        self.verticalHeader().setVisible(True)
        self.horizontalHeader().setDefaultAlignment(Qt.AlignCenter)
        self.horizontalHeader().setStretchLastSection(True)
        #the built-in vertical scrollBar of this view is always off
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.installEventFilter(self)

    def setVerticalScrollBar(self, vScrollBar):
        self.vScrollBar = vScrollBar
        self.vScrollBar.valueChanged.connect(self.onScrollbarValueChanged)
        self.vScrollBar.setVisible(False)

    def setModel(self, model):
        super().setModel(model)
        model.rowCountChanged.connect(self.onRowCountChanged)
        model.beginViewPortRefresh.connect(self.onBeginViewportRefresh)
        model.endViewPortRefresh.connect(self.onEndViewportRefresh)
        self.horizontalHeader().sortIndicatorChanged.disconnect()
        self.setSortingEnabled(False)

    def setTrackingColumn(self, col):
        """column used to track a selected row while scrolling on this table"""
        self.trackingCol = col

    def getRowCells(self, row):
        cols = []
        for i in range(0, self.model().columnCount()):
            c = self.model().index(row, i)
            cols.append(c.data())
        return cols

    def clear(self):
        pass

    def refresh(self):
        self.calculateRowsInViewport()
        self.model().setRowCount(min(self.maxRowsInViewport, self.model().totalRowCount))
        self.model().refreshViewport(self.vScrollBar.value(), self.maxRowsInViewport, force=True)

    def forceViewRefresh(self):
        return (self.vScrollBar.minimum() == self.vScrollBar.value() or self.vScrollBar.maximum() == self.vScrollBar.value())

    def calculateRowsInViewport(self):
        rowHeight = self.verticalHeader().defaultSectionSize()
        # we don't want partial-height rows in viewport, hence .floor()
        self.maxRowsInViewport = math.floor(self.viewport().height() / rowHeight)+1

    def scrollViewport(self, row):
        maxVal = self.maxRowsInViewport-1
        if row >= maxVal:
            pos = self.vScrollBar.value() + 1
            self.vScrollBar.setValue(pos)
            return pos
        elif row == 0:
            pos = max(0, self.vScrollBar.value() - 1)
            self.vScrollBar.setValue(pos)
            return pos

        return None

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)
        self.mousePressed = False
        if event.button() != Qt.LeftButton:
            return

        for idx in self.selectionModel().selectedRows(self.trackingCol):
            if idx.data() != None and idx.data() not in self._rows_selection.keys():
                self._rows_selection[idx.data()] = self.getRowCells(idx.row())

        # TODO: handle selection ranges when Shift is pressed
        self._selectSavedIndex()

    def mouseMoveEvent(self, event):
        try:
            pos = event.pos()
            row = self.rowAt(pos.y())
            item = self.indexAt(pos)

            if item == None:
                return

            clickedItem = self.model().index(row, self.trackingCol)
            if clickedItem.data() == None:
                return
            self.handleMouseMoveEvent(row, clickedItem, self.selectionModel().isRowSelected(row, QModelIndex()))

        finally:
            # call upper implementation to select/deselect rows.
            super().mouseMoveEvent(event)

    def handleMouseMoveEvent(self, row, clickedItem, selected):
        if not selected:
            if clickedItem.data() in self._rows_selection.keys():
                del self._rows_selection[clickedItem.data()]
        else:
            self._rows_selection[clickedItem.data()] = self.getRowCells(row)

        # handle scrolling the view while dragging the mouse.
        if self.mousePressed:
            scrollPos = self.scrollViewport(row)
            if scrollPos == None:
                return

            nextItem = self.model().index(scrollPos, self.trackingCol)
            if nextItem == None or nextItem.data() == None:
                return
            if clickedItem.data() not in self._rows_selection.keys():
                self._rows_selection[nextItem.data()] = self.getRowCells(nextItem.row())

    # save the selected index, to preserve selection when moving around.
    def mousePressEvent(self, event):
        # we need to call upper class to paint selections properly
        super().mousePressEvent(event)
        rightBtnPressed = event.button() != Qt.LeftButton

        pos = event.pos()
        item = self.indexAt(pos)
        row = self.rowAt(pos.y())
        if item == None:
            return

        clickedItem = self.model().index(row, self.trackingCol)
        if clickedItem.data() == None:
            return

        self.mousePressed = not rightBtnPressed
        flags = QItemSelectionModel.Rows | QItemSelectionModel.SelectCurrent

        # 1. if ctrl is pressed, select / deselect current row
        # 2. if ctrl is not pressed:
        #   1. discard previous selection
        #   2. select current line if it's not selected. Deselect it otherwise.
        # 3. if ctrl is not pressed and there's more than one row selected, and
        # the clicked row is selected: discard selection, and select current
        # clicked row.
        # 4. if Left button has not been pressed, do not discard the selection.
        rowSelected = clickedItem.data() in self._rows_selection.keys()
        if self.ctrlPressed:
            if rowSelected:
                del self._rows_selection[clickedItem.data()]
                flags = QItemSelectionModel.Rows | QItemSelectionModel.Deselect
        else:
            deselectCurRow = len(self._rows_selection.keys()) == 1 and not rightBtnPressed
            # discard current selection:
            # - if the user right clicked on a row not part of a selection.
            # - if the user clicked on a row, and there's only one row
            # selected.
            # - if the user clicked on a row already selected.
            if (not rowSelected and rightBtnPressed) or not rightBtnPressed or deselectCurRow:
                self.selectionModel().clear()
                self._rows_selection = {}
            if rowSelected and deselectCurRow and not rightBtnPressed:
                flags = QItemSelectionModel.Rows | QItemSelectionModel.Deselect
            else:
                self._rows_selection[clickedItem.data()] = self.getRowCells(row)

        self.selectionModel().setCurrentIndex(
            clickedItem,
            flags
        )

    def onBeginViewportRefresh(self):
        # if the selected row due to scrolling up/down doesn't match with the
        # saved index, deselect the row, because the saved index is out of the
        # view.
        pass

    def onEndViewportRefresh(self):
        if not self.mousePressed and not self.shiftPressed:
            self.selectionModel().clear()
        self._selectSavedIndex()

    def resizeEvent(self, event):
        super().resizeEvent(event)
        #refresh the viewport data based on new geometry
        self.refresh()

    def onRowCountChanged(self):
        totalCount = self.model().totalRowCount
        self.vScrollBar.setVisible(True if totalCount > self.maxRowsInViewport else False)

        self.vScrollBar.setMinimum(0)
        # we need to substract the displayed rows to the total rows, to scroll
        # down correctly.
        self.vScrollBar.setMaximum(max(0, totalCount - self.maxRowsInViewport+1))

        self.model().refreshViewport(self.vScrollBar.value(), self.maxRowsInViewport, force=self.forceViewRefresh())

    def clearSelection(self):
        self.selectionModel().reset()
        self.selectionModel().clearCurrentIndex()

    def selectedRows(self, limit=""):
        model = self.selectionModel()
        curModel = self.model()
        selection = model.selectedRows()
        if not selection:
            return None

        rows = []
        for k in self._rows_selection:
            rows.append(self._rows_selection[k])
        return rows

    def getCurrentIndex(self):
        return self.selectionModel().currentIndex().internalId()

    def selectItem(self, _data, _column):
        """Select a row based on the data displayed on the given column.
        """
        items = self.model().findItems(_data, column=_column)
        if len(items) > 0:
            self.selectionModel().setCurrentIndex(
                items[0].index(),
                QItemSelectionModel.Rows | QItemSelectionModel.SelectCurrent
            )

    def _selectSavedIndex(self):
        sel = QItemSelection()

        for text in self._rows_selection.keys():
            items = self.model().findItems(text, column=self.trackingCol)
            if len(items) == 0:
                continue

            for i in items:
                sel.append(QItemSelectionRange(i.index()))
            self.selectionModel().select(sel, QItemSelectionModel.Select | QItemSelectionModel.Rows)

    def _selectLastRow(self):
        internalId = self.getCurrentIndex()
        self.selectionModel().setCurrentIndex(
            self.model().createIndex(self.maxRowsInViewport-2, self.trackingCol, internalId),
            QItemSelectionModel.Rows | QItemSelectionModel.SelectCurrent
        )

    def _selectRow(self, pos):
        internalId = self.getCurrentIndex()
        self.selectionModel().setCurrentIndex(
            self.model().createIndex(pos, self.trackingCol, internalId),
            QItemSelectionModel.Rows | QItemSelectionModel.SelectCurrent
        )

    def onScrollbarValueChanged(self, vSBNewValue):
        self.model().refreshViewport(vSBNewValue, self.maxRowsInViewport, force=True)

    def onKeyUp(self):
        curIdx = self.selectionModel().currentIndex()
        if not self.shiftPressed:
            self._rows_selection = {}
        self._rows_selection[curIdx.data()] = self.getRowCells(curIdx.row())

        if self.selectionModel().currentIndex().row() == 0:
            self.vScrollBar.setValue(max(0, self.vScrollBar.value() - 1))

    def onKeyDown(self):
        curIdx = self.selectionModel().currentIndex()
        curRow = curIdx.row()
        if not self.shiftPressed:
            self._rows_selection = {}
        self._rows_selection[curIdx.data()] = self.getRowCells(curRow)

        if curRow >= self.maxRowsInViewport-2:
            self.onKeyPageDown()
            self._selectRow(0)
        else:
            self._selectRow(curRow)

    def onKeyHome(self):
        self.vScrollBar.setValue(0)
        if not self.mousePressed and self.shiftPressed:
            self.selectionModel().clear()

    def onKeyEnd(self):
        self.vScrollBar.setValue(self.vScrollBar.maximum())
        if not self.mousePressed and self.shiftPressed:
            self.selectionModel().clear()
        self._selectLastRow()

    def onKeyPageUp(self):
        newValue = max(0, self.vScrollBar.value() - self.maxRowsInViewport)
        self.vScrollBar.setValue(newValue)

    def onKeyPageDown(self):
        if self.vScrollBar.isVisible() == False:
            return

        newValue = self.vScrollBar.value() + (self.maxRowsInViewport-2)
        self.vScrollBar.setValue(newValue)

    def onKeySpace(self):
        if self.vScrollBar.isVisible() == False:
            return

        newValue = self.vScrollBar.value() + (self.maxRowsInViewport-2)
        self.vScrollBar.setValue(newValue)

    def eventFilter(self, obj, event):
        if event.type() == QEvent.KeyRelease:
            if event.key() == Qt.Key_Shift:
                self.shiftPressed = False
            if event.key() == Qt.Key_Control:
                self.ctrlPressed = False

        elif event.type() == QEvent.KeyPress:
            # FIXME: setValue() does not update the scrollbars correctly in
            # some pyqt versions.
            if event.key() == Qt.Key_Up:
                self.onKeyUp()
            elif event.key() == Qt.Key_Down:
                self.onKeyDown()
            elif event.key() == Qt.Key_Home:
                self.onKeyHome()
            elif event.key() == Qt.Key_End:
                self.onKeyEnd()
            elif event.key() == Qt.Key_PageUp:
                self.onKeyPageUp()
            elif event.key() == Qt.Key_PageDown:
                self.onKeyPageDown()
            elif event.key() == Qt.Key_Escape:
                self.selectionModel().clear()
                self._rows_selection = {}
            elif event.key() == Qt.Key_Shift:
                self.shiftPressed = True
            elif event.key() == Qt.Key_Control:
                self.ctrlPressed = True
            elif event.key() == Qt.Key_Space:
                self.onKeySpace()

        elif event.type() == QEvent.Wheel:
            self.vScrollBar.wheelEvent(event)
            return True

        return super(GenericTableView, self).eventFilter(obj, event)



import logging
import requests
import json
import re

from PyQt5 import QtCore
from opensnitch.version import version
from opensnitch.config import Config
from opensnitch.plugins import PluginBase, PluginSignal
from opensnitch.dialogs.prompt import PromptDialog
from opensnitch.dialogs.processdetails import ProcessDetailsDialog
from opensnitch.plugins.virustotal import _popups
from opensnitch.plugins.virustotal import _procdialog

ch = logging.StreamHandler()
#ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - [%(levelname)s] %(message)s')
ch.setFormatter(formatter)
logger = logging.getLogger(__name__)
logger.addHandler(ch)
logger.setLevel(logging.WARNING)

class VTSignals(QtCore.QObject):
    completed = QtCore.pyqtSignal(object, object, object, object, object)
    error = QtCore.pyqtSignal(str, object, object, str, object)

class VTAnalysis(QtCore.QRunnable):
    def __init__(self, parent, config, what, url, timeout, api_key, conn):
        super(VTAnalysis, self).__init__()
        #QtCore.QThread.__init__(self)
        self.signals = VTSignals()
        self.parent = parent
        self.config = config
        self.what = what
        self.url = url
        self.timeout = timeout
        self.api_key = api_key
        self.conn = conn

    def run(self):
        self.analyze(self.parent, self.config, self.what, self.url, self.timeout, self.api_key, self.conn)

    def analyze(self, parent, conf, what, url, timeout, api_key, conn):
        """Returns analyze results on success.
        None on error.
        """
        logger.debug("vt_analysis.analyze() %s, %s, %s", url, what, type(parent))
        response={}
        try:
            # >= v2.13. Not tested with < v2.13
            response = requests.get(
                url,
                timeout=timeout,
                headers={'x-apikey': api_key, 'User-Agent': 'OpenSnitch v%s' % version}
            )
            self.signals.completed.emit(what, parent, conf, conn, response)
        except Exception as e:
            logger.warning("vt_analysis.analyze() exception: %s", repr(e))
            self.signals.error.emit(what, parent, conf, "Exception: {0}".format(e), response)

class Virustotal(PluginBase):
    """
    Analyzes properties of a connection: domain, IP, file hash.
    json format:
        {
            "config": {
              "api_timeout": 2,
              "api_key": "123456",
              "api_domains_url": "https://www.virustotal.com/api/v3/domains/",
              "api_ips_url": "https://www.virustotal.com/api/v3/ip_addresses/",
              "api_files_url": "https://www.virustotal.com/api/v3/files/"
            },
            "check": ["domains", "ips", "files"]
        }

    The config item is optional, and will override default configuration if specified.

    Documentation:
     - get an API key: https://virustotal.readme.io/docs/please-give-me-an-api-key
     - https://developers.virustotal.com/reference/domain-info
     - https://developers.virustotal.com/reference/domains-1
     - https://developers.virustotal.com/reference/ip-info
     - https://developers.virustotal.com/reference/file-info


     Privacy warning: remember that the domains, ips or hashes will be sent to a remote
    """
    # Friendly reminder: this is just a PoC, an example of what can be done.

    # XXX ideas:
    # - Send a desktop notification if something is detected as malicious.
    # - [DONE] Add a tab page to the popup with the result of the analysis.
    #   * Add a link to know more about the domain, certificates, etc.
    # - Change the icon if malicious.
    # - Hook Procs tab list, get the hash of each binary, upload to VT and
    # update the list with info.
    # - From the process dialog, allow to upload the binary, and obtain the process
    # analysis and behaviour.
    #   * [DONE] added tab with info about the process
    # - Add a button to upload a file for analysis.
    # - Periodically scan all domains/IPs visited.
    #   * from procs/hosts/ips -> on demand.
    #   * as soon as we add a domain to the DB?

    # TODO: similar services
    # urlhaus-api.abuse.ch
    #
    # analysis with yara rules: yaraify-api.abuse.ch
    # curl -X POST -d '{ "query": "lookup_hash", "search_term": "38e1c0ca15ed83ed27148c31a31e0b33de627519ab2929d4aa69484534589086" '} https://yaraify-api.abuse.ch/api/v1/

    # malware domain test: malware.wicar.org

    name = "Virustotal"
    version = 0
    author = "opensnitch"
    created = ""
    modified = ""
    #enabled = False
    description = "Analyze domains and IPs with VirusTotal"

    # where this plugin is allowed
    # could be applied on list of connections, process dialog, etc
    TYPE = [PluginBase.TYPE_POPUPS]

    # result of the query
    QUERY_OK = 0
    QUERY_ERROR = 1

    # verdict of the analysis
    VERDICT_UNKNOWN = 0
    VERDICT_BENIGN = 1
    VERDICT_MALICIOUS = 2

    ANALYZING_MESSAGE = "\u231B Virustotal: analyzing ..."
    VERDICT_BENIGN_MESSAGE = "\u2714 Virustotal"
    VERDICT_MALICIOUS_MESSAGE = "\u26A0  Virustotal warning"

    # Virustotal returns the analysis of many engines (> 70)
    # Sometimes only 1 analysis returns malicious, which may lead to false
    # positives.

    # consider the object (domain, ip, etc) malicious if the number of reports
    # is equal or above this threshold.
    MALICIOUS_THRESHOLD = 1
    WARNING_THRESHOLD = 5
    # There's also a community reputation:
    # reputation: <integer> domain's score calculated from the votes of the
    # VirusTotal's community.

    API_DOMAINS = "https://www.virustotal.com/api/v3/domains/"
    API_IPS = "https://www.virustotal.com/api/v3/ip_addresses/"
    #sha256, sha1 or md5
    API_FILES = "https://www.virustotal.com/api/v3/files/"

    API_EXCEEDED = False
    API_KEY = 'https://virustotal.readme.io/docs/please-give-me-an-api-key'
    API_CONNECT_TIMEOUT = 1

    # urls to view the details of an object
    RESULTS_DOMAINS = "https://www.virustotal.com/gui/domain/"
    RESULTS_IPS = "https://www.virustotal.com/gui/ip-address/"

    CHECK_DOMAINS = "domains"
    CHECK_IPS = "ips"
    CHECK_FILES_HASHES = "hashes"
    # TODO
    # CHECK_PROC_BEHAVIOUR = "behaviour"

    classA_net = r'10\.\d{1,3}\.\d{1,3}\.\d{1,3}'
    classB_net = r'172\.1[6-9]\.\d+\.\d+|172\.2[0-9]\.\d+\.\d+|172\.3[0-1]+\.\d{1,3}\.\d{1,3}'
    classC_net = r'192\.168\.\d{1,3}\.\d{1,3}'
    others_net = r'127\.\d{1,3}\.\d{1,3}\.\d{1,3}|169\.254\.\d{1,3}\.\d{1,3}'
    multiIPv4 = r'2[32][23459]\.\d{1,3}\.\d{1,3}\.\d{1,3}'
    multiIPv6 = r'ffx[0123458ef]::'
    MULTICAST_RANGE = "^(" + multiIPv4 + ")$"
    LAN_RANGES = "^(" + others_net + "|" + classC_net + "|" + classB_net + "|" + classA_net + multiIPv4 + "|" + multiIPv6 + "|::1|f[cde].*::.*)$"

    def __init__(self, config=None):
        self.API_DOMAINS = "https://www.virustotal.com/api/v3/domains/"
        self.API_IPS = "https://www.virustotal.com/api/v3/ip_addresses/"
        #sha256, sha1 or md5
        self.API_FILES = "https://www.virustotal.com/api/v3/files/"

        self.API_EXCEEDED = False
        self.API_KEY = ''
        self.API_CONNECT_TIMEOUT = 1

        self.MALICIOUS_THRESHOLD = 1
        self.WARNING_THRESHOLD = 5

        # original json config received
        self._config = config
        self.ip_regex = re.compile(r"^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$")
        self.lan_regex = re.compile(Virustotal.LAN_RANGES)
        self.signal_in.connect(self.cb_incoming_events)
        self.threadsPool = QtCore.QThreadPool()

    def configure(self, parent=None):
        """add widgets to all supported areas of the GUI"""

        if type(parent) == PromptDialog:
            vt_tab = _popups.build_vt_tab(self, parent)
            _popups.add_vt_tab(parent, vt_tab)

        elif type(parent) == ProcessDetailsDialog:
            vt_tab = _procdialog.build_vt_tab(self, parent)
            _procdialog.add_vt_tab(self, parent, vt_tab)

    def compile(self):
        """Transform json items to python objects, if needed.
        It's executed only once.
        """
        logger.debug("compile()")
        for idx in self._config:
            if idx == "config" and 'api_timeout' in self._config[idx]:
                self.API_CONNECT_TIMEOUT = self._config[idx]['api_timeout']
            if idx == "config" and 'api_key' in self._config[idx]:
                self.API_KEY = self._config[idx]['api_key']
            if idx == "config" and 'api_domains_url' in self._config[idx]:
                self.API_DOMAINS = self._config[idx]['api_domains_url']
            if idx == "config" and 'api_ips_url' in self._config[idx]:
                self.API_IPS = self._config[idx]['api_ips_url']
            if idx == "config" and 'api_files_url' in self._config[idx]:
                self.API_FILES = self._config[idx]['api_files_url']

        if self._config.get("malicious-label-style") == None:
            self._config['malicious-label-style'] = 'red'
        if self._config.get("benign-label-style") == None:
            self._config['benign-label-style'] = 'green'
        if self._config.get("malicious-message") == None:
            self._config['malicious-message'] = self.VERDICT_MALICIOUS_MESSAGE
        if self._config.get("benign-message") == None:
            self._config['benign-message'] = self.VERDICT_BENIGN_MESSAGE

    def run(self, parent, args):
        """
        arg0 - object: parent object,
        arg1 - list: connection,

        Pre-requisites: create a rule to allow outbound connections to
        www.virustotal.com on port 443 (from your uid).
        """
        # parent == PromptDialog
        self.parent = parent

        try:
            if type(parent) == PromptDialog:
                parent.messageLabel.linkActivated.connect(lambda link: _popups._cb_popup_link_clicked(link, parent))
                _popups.reset_widgets_state(parent)
                #_popups.add_analyzing_msg(self, parent)
                conn = args[0]
                if 'www.virustotal.com' == conn.dst_host:
                    return
                if self.lan_regex.match(conn.dst_host) != None or self.lan_regex.match(conn.dst_ip) != None:
                    return

                logger.debug("analyzing %s, %s", self.API_DOMAINS, conn.dst_host)

                # parse config file "virustotal": {}
                url = ""
                for what in self._config['check']:
                    if Virustotal.CHECK_DOMAINS == what:
                        if conn.dst_host == "" or conn.dst_ip == conn.dst_host:
                            continue
                        url = self.API_DOMAINS + conn.dst_host
                    elif Virustotal.CHECK_IPS == what:
                        url = self.API_IPS + conn.dst_ip
                    elif Virustotal.CHECK_FILES_HASHES == what:
                        checksum = conn.process_checksums[Config.OPERAND_PROCESS_HASH_MD5]
                        if checksum != "":
                            url = self.API_FILES + checksum
                        else:
                            logger.debug("run() checksum of this process empty, skipping")
                            continue
                    else:
                        logger.info("run() unknown target: %s", what)
                        continue

                    logger.debug("run() analyzing: %s", url)

                    vt_thread = VTAnalysis(parent, self._config, what, url, self.API_CONNECT_TIMEOUT, self.API_KEY, conn)
                    vt_thread.signals.completed.connect(self.analysis_completed)
                    vt_thread.signals.error.connect(self.analysis_error)
                    self.threadsPool.start(vt_thread)

            else:
                print("Virustotal error: parent type not supported:", type(parent))
        except Exception as e:
            logger.warning("run() exception: %s", repr(e))
        #finally:
        #    if type(parent) == PromptDialog:
        #        parent.stackedWidget.removeWidget(vt_tab)

    def analysis_error(self, what, parent, conf, error, response):
        logger.warning("analysis_error(): %s, %s, %s", what, error, repr(response))
        if type(parent) == PromptDialog:
            self.update_popup(what, response, parent, conf, error)

    def analysis_completed(self, what, parent, conf, conn, response):
        try:
            self.API_EXCEEDED = (response.status_code == 403 or response.status_code == 204)
            if self.API_EXCEEDED:
                logger.info("analysis_completed() API USAGE EXCEEDED")
                return

            if type(parent) == PromptDialog:
                self.update_popup(what, response, parent, conf)
            elif type(parent) == ProcessDetailsDialog:
                _procdialog.update_tab(what, response, parent, conf, conn)
            else:
                print("[virustotal] analysis_completed() parent object type not supported:", type(parent))

        except Exception as e:
            logger.warning("analysis_completed() exception: %s", repr(e))
            return

    def update_popup(self, what, response, parent, config, errmsg=None):
        """Update pop-up widgets based on the results of the analysis.
        """
        # XXX: use PromptDialog methods to update widgets
        # set_app_description()
        # set_app_path() <- allow to colorize text
        # etc.

        error = (errmsg != None)
        malicious = False
        labelStyle = "color: {0}".format(config['benign-label-style'])
        try:
            if error:
                if response and response.content == 401:
                    raise ValueError("Unauthorized (401).\nCheck the validity of the Virustotal API key.\n\n{0}".format(errmsg))
                else:
                    raise ValueError(errmsg)
            #if response.get('content') == None:
            #    raise Exception("Invalid response from server?")

            result = json.loads(response.content)
            # checksums API returns 404 if the hash is not in the DB.
            # XXX: result of this query could be marked as 'unknown' instead of
            # 'benign'.
            if result.get('data') == None:
                logger.debug("update_popup() no data? %s, %s", what, response)
                if Virustotal.CHECK_FILES_HASHES == what:
                    return

            verdict = result['data']['attributes']['last_analysis_stats']
            #print("[Virustotal] RESULT:\n", conn.dst_host, "\n", result['data']['attributes']['last_analysis_stats'])

            # XXX: if we analyze multiple objects (domains, ips, hashes...),
            # onlye the last response is stored.
            _popups.add_vt_response(parent, result)

            malicious = self.is_malicious(verdict['malicious'])

            # self.set_malicious(parent)
        except ValueError as e:
            logger.warning("update_popup() value error: %s -> %s", repr(e), response)
            error = True
            _popups.add_vt_response(parent, None, e)
        except Exception as e:
            logger.warning("update_popup() exception: %s -> %s", repr(e), response)
            error = True
            _popups.add_vt_response(parent, None, "Exception: {0}".format(repr(e)))

        finally:
            message = "<font color=\"{0}\">{1} ({2})</font><br>{3}".format(
                config['benign-label-style'],
                config['benign-message'],
                what,
                parent.messageLabel.text()
            )
            if malicious:
                labelStyle = "color: {0}".format(config['malicious-label-style'])
                message = "<font color=\"{0}\">{1} ({2}, flagged by {3} sources)</font> <a href='#virustotal-warning'>(Details)</a><br>{4}".format(
                    config['malicious-label-style'],
                    config['malicious-message'],
                    what,
                    verdict['malicious'],
                    parent.messageLabel.text()
                )
                parent.messageLabel.setStyleSheet(labelStyle)
            if error:
                labelStyle = "color: darkOrange"
                message = "<font color=\"darkOrange\">Virustotal ({0}): analysis error</font> <a href='#virustotal-warning'>(Details)</a><br>{1}".format(
                    what,
                    parent.messageLabel.text()
                )
                parent.messageLabel.setStyleSheet(labelStyle)


            parent.messageLabel.setText(message)

            if Virustotal.CHECK_DOMAINS == what:
                parent.messageLabel.setStyleSheet(labelStyle)
            if Virustotal.CHECK_IPS == what:
                parent.destIPLabel.setStyleSheet(labelStyle)
            if Virustotal.CHECK_FILES_HASHES == what:
                parent.checksumLabel.setStyleSheet(labelStyle)

    def is_benign(self, mal_results):
        return mal_results < self.MALICIOUS_THRESHOLD

    def is_malicious(self, mal_results):
        return mal_results >= self.MALICIOUS_THRESHOLD

    def cb_incoming_events(self, signal):
        """listens to events from parents (enable, configure, etc)
        """
        logger.debug("cb_incoming_events() %s", signal)
        try:
            if signal['signal'] == PluginSignal.ENABLE:
                self.enabled = True
            if signal['signal'] == PluginSignal.DISABLE:
                self.enabled = False
                #if not self.enabled:
                #    _popups.remove_vt_tab()
                #    _procdialog.remove_vt_tab()
        except Exception as e:
            logger.debug("cb_incoming_events() exception: %s", repr(e))
